import ctypes as ct
from monero_glue.xmr.core.backend import trezor_types as tt

# 1. regex the Header file to CLIB.METHOD.argtypes=[]
# 2. replace the argtypes wiht our types
# 3. regex the method list to the buff list of methods
# 4. put return value as the first element if the return elem is not None
# CLIB\.(.+?)\.argtypes\s*=\s*\[(.+?)\]


buff = [
    ['curve25519_copy', tt.FE, tt.FE],
    ['curve25519_add', tt.FE, tt.FE, tt.FE],
    ['curve25519_add_after_basic', tt.FE, tt.FE, tt.FE],
    ['curve25519_add_reduce', tt.FE, tt.FE, tt.FE],
    ['curve25519_sub', tt.FE, tt.FE, tt.FE],
    ['curve25519_scalar_product', tt.FE, tt.FE, ct.c_uint32],
    ['curve25519_sub_after_basic', tt.FE, tt.FE, tt.FE],
    ['curve25519_sub_reduce', tt.FE, tt.FE, tt.FE],
    ['curve25519_neg', tt.FE, tt.FE],
    ['curve25519_mul', tt.FE, tt.FE, tt.FE],
    ['curve25519_square', tt.FE, tt.FE],
    ['curve25519_square_times', tt.FE, tt.FE, ct.c_int],
    ['curve25519_expand', tt.FE, tt.KEY_BUFF],
    ['curve25519_contract', tt.KEY_BUFF, tt.FE],
    ['curve25519_recip', tt.FE, tt.FE],

    ['curve25519_set', tt.FE, ct.c_uint32],
    [ct.c_int, 'curve25519_isnegative', tt.FE],
    [ct.c_int, 'curve25519_isnonzero', tt.FE],
    ['curve25519_reduce', tt.FE, tt.FE],
    ['curve25519_expand_reduce', tt.FE, tt.KEY_BUFF],

    [ct.c_int, 'ed25519_verify', ct.c_char_p, ct.c_char_p, ct.c_size_t],
    ['ge25519_p1p1_to_partial', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_p1p1)],
    ['ge25519_p1p1_to_full', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_p1p1)],
    ['ge25519_full_to_pniels', ct.POINTER(tt.Ge25519_pniels), ct.POINTER(tt.Ge25519)],

    ['ge25519_double_p1p1', ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519)],
    ['ge25519_nielsadd2_p1p1', ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_niels), ct.c_char],
    ['ge25519_pnielsadd_p1p1', ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_pniels), ct.c_char],
    ['ge25519_double_partial', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_double', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_nielsadd2', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_niels)],
    ['ge25519_pnielsadd', ct.POINTER(tt.Ge25519_pniels), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_pniels)],
    ['ge25519_pack', tt.KEY_BUFF, ct.POINTER(tt.Ge25519)],
    [ct.c_int, 'ge25519_unpack_negative_vartime', ct.POINTER(tt.Ge25519), tt.KEY_BUFF],
    ['ge25519_set_neutral', ct.POINTER(tt.Ge25519)],
    ['ge25519_double_scalarmult_vartime', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM, tt.MODM],
    ['ge25519_scalarmult', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM],
    [ct.c_int, 'ge25519_check', ct.POINTER(tt.Ge25519)],
    [ct.c_int, 'ge25519_eq', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_copy', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_set_base', ct.POINTER(tt.Ge25519)],
    ['ge25519_mul8', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_neg_partial', ct.POINTER(tt.Ge25519)],
    ['ge25519_neg_full', ct.POINTER(tt.Ge25519)],
    ['ge25519_reduce', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_norm', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)],
    ['ge25519_add', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.c_char],
    ['ge25519_fromfe_frombytes_vartime', ct.POINTER(tt.Ge25519), tt.KEY_BUFF],
    [ct.c_int, 'ge25519_unpack_vartime', ct.POINTER(tt.Ge25519), tt.KEY_BUFF],
    ['ge25519_scalarmult_base_wrapper', ct.POINTER(tt.Ge25519), tt.MODM],
    ['ge25519_scalarmult_wrapper', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM],
    ['ge25519_set_xmr_h', ct.POINTER(tt.Ge25519)],


    ['xmr_random_scalar', tt.MODM],
    ['xmr_fast_hash', tt.KEY_BUFF, ct.c_void_p, ct.c_size_t],

    ['xmr_hash_to_scalar', tt.MODM, ct.c_void_p, ct.c_size_t],
    ['xmr_hash_to_ec', ct.POINTER(tt.Ge25519), ct.c_void_p, ct.c_size_t],
    ['xmr_derivation_to_scalar', tt.MODM, ct.POINTER(tt.Ge25519), ct.c_uint32],
    ['xmr_generate_key_derivation', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM],
    ['xmr_derive_private_key', tt.MODM, ct.POINTER(tt.Ge25519), ct.c_uint32, tt.MODM],
    ['xmr_derive_public_key', ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.c_uint32, ct.POINTER(tt.Ge25519)],
    ['xmr_gen_c', ct.POINTER(tt.Ge25519), tt.MODM, ct.c_uint64],
    ['xmr_add_keys1', ct.POINTER(tt.Ge25519), tt.MODM, tt.MODM, ct.POINTER(tt.Ge25519)],
    ['xmr_add_keys2', ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519)],
    ['xmr_get_subaddress_secret_key', tt.MODM, ct.c_uint32, ct.c_uint32, tt.MODM],
]


args = ['r', 'a', 'b', 'c', 'd', 'e', 'f', 'h']
for xx in buff:
    ret_arg = xx[0] if not isinstance(xx[0], str) else None
    arg_off = 0
    if ret_arg is not None:
        xx = xx[1:]
        arg_off = 1

    arg_len = len(xx[1:])
    arg_str = ', '.join(args[arg_off:arg_len+arg_off])
    cargs = xx[1:]
    arg_par = ', '.join([('ct.byref(%s)' % args[ii+arg_off]) if isinstance(x, type(ct.POINTER(tt.Ge25519))) else args[ii+arg_off] for ii, x in enumerate(cargs)])
    tpl = 'def %s(%s): \n' % (xx[0], arg_str)
    tpl += '    return CLIB.%s(%s)\n' % (xx[0], arg_par)
    tpl += '\n'
    print(tpl)

    # _r version guess
    if ret_arg is not None:
        continue

    arg_off += 1
    arg_len = len(xx[2:])  # assume first param is output
    arg_str = ', '.join(args[arg_off:arg_len + arg_off])
    cargs = xx[1:]
    arg_par = ', '.join(
        [('ct.byref(%s)' % args[ii + arg_off-1]) if isinstance(x, type(ct.POINTER(tt.Ge25519))) else args[ii + arg_off-1]
         for ii, x in enumerate(cargs)])

    rarg = None
    rarg_type = cargs[0]
    if isinstance(cargs[0], type(ct.POINTER(tt.Ge25519))):
        rarg_type = cargs[0]._type_

    if rarg_type == tt.Ge25519:
        rarg = 'tt.Ge25519()'
    elif rarg_type == tt.Ge25519_pniels:
        rarg = 'tt.Ge25519_pniels()'
    elif rarg_type == tt.Ge25519_niels:
        rarg = 'tt.Ge25519_niels()'
    elif rarg_type == tt.Ge25519_p1p1:
        rarg = 'tt.Ge25519_p1p1()'
    elif rarg_type == tt.MODM:
        rarg = 'tt.MODM()'
    elif rarg_type == tt.FE:
        rarg = 'tt.FE()'

    tpl = 'def %s_r(%s): \n' % (xx[0], arg_str)
    tpl += '    r = %s\n' % (rarg if rarg is not None else ' None  # TODO: fix')
    tpl += '    CLIB.%s(%s)\n' % (xx[0], arg_par)
    tpl += '    return r\n'
    tpl += '\n'
    print(tpl)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest
import binascii
import logging

import aiounittest

from monero_glue.xmr import crypto, common
from monero_glue.xmr.core import ec_py


logger = logging.getLogger(__name__)


try:
    from monero_glue.xmr.core import ec_trezor
    tcry = ec_trezor.tcry
    LOADED = 1

except Exception as e:
    logger.info('Trezor backend loading error: %s' % e)
    LOADED = 0


class EcTrezorTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(EcTrezorTest, self).__init__(*args, **kwargs)

    def test_ed_crypto(self):
        if not LOADED:
            self.skipTest('Trezor crypto missing')
            
        h_hex = b'8b655970153799af2aeadc9ff1add0ea6c7251d54154cfa92c173a0dd39c1f94'
        h = binascii.unhexlify(h_hex)
        pt = tcry.ge25519_unpack_vartime_r(tcry.KEY_BUFF(*bytes(h)))
        packed = tcry.ge25519_pack_r(pt)
        self.assertEqual(h, packed)

    def test_modm(self):
        if not LOADED:
            self.skipTest('Trezor crypto missing')

        x = ec_py.encodeint(0xaa)
        y = ec_trezor.decodeint(x)
        yy = ec_trezor.encodeint(y)
        self.assertEqual(yy[0], 0xaa)
        self.assertEqual(yy[1], 0x00)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018


import logging
from monero_glue.xmr import crypto

logger = logging.getLogger(__name__)


def gen_schnorr_non_linkable(x, P1, P2, index):
    """
    Generates simple Schnorr
    :param x:
    :param P1:
    :param P2:
    :param index:
    :return:
    """
    a = crypto.random_scalar()
    L1 = crypto.scalarmult_base(a)
    s2 = crypto.random_scalar()
    c2 = crypto.cn_fast_hash(crypto.encodepoint(L1))
    L2 = crypto.point_add(crypto.scalarmult_base(s2), crypto.scalarmult(P2 if index == 0 else P1, crypto.decodeint(c2)))
    c1 = crypto.cn_fast_hash(crypto.encodepoint(L2))
    s1 = crypto.sc_mulsub(x, crypto.decodeint(c1), a)

    return (L1, s1, s2) if index == 0 else (L2, s2, s1)


def ver_schnorr_non_linkable(P1, P2, L1, s1, s2):
    """
    Verifies Schnorr signature generated by gen_schnorr_non_linkable
    :param P1:
    :param P2:
    :param L1:
    :param s1:
    :param s2:
    :return:
    """
    c2 = crypto.cn_fast_hash(crypto.encodepoint(L1))
    L2 = crypto.point_add(crypto.scalarmult_base(s2), crypto.scalarmult(P2, crypto.decodeint(c2)))
    c1 = crypto.cn_fast_hash(L2)
    L1p = crypto.point_add(crypto.scalarmult_base(s1), crypto.scalarmult(P1, crypto.decodeint(c1)))

    if L1 == L1p:
        return 0

    else:
        logger.warning('Didn\'t verify L1: %s, L1p: %s' % (L1, L1p))
        return -1
    

def gen_asnl(x, P1, P2, indices):
    """
    Aggregate Schnorr Non-Linkable
    x, P1, P2, are key vectors here, but actually you
    indices specifices which column of the given row of the key vector you sign.
    the key vector with the first or second key

    :param x:
    :param P1:
    :param P2:
    :param indices:
    :return:
    """
    n = len(x)
    logger.info("Generating Aggregate Schnorr Non-linkable Ring Signature")

    L1 = [None] * n
    s1 = [None] * n
    s2 = [None] * n
    s = crypto.sc_0()
    for j in range(0, n):
        L1[j], s1[j], s2[j] = gen_schnorr_non_linkable(x[j], P1[j], P2[j], indices[j])
        s = crypto.sc_add(s, s1[j])
    return L1, s2, s

        
def ver_asnl(P1, P2, L1, s2, s):
    """
    Aggregate Schnorr Non-Linkable
    :param P1:
    :param P2:
    :param L1:
    :param s2:
    :param s:
    :return:
    """
    logger.info("Verifying Aggregate Schnorr Non-linkable Ring Signature")

    n = len(P1)
    LHS = crypto.scalarmult_base(crypto.sc_0())
    RHS = crypto.scalarmult_base(s)
    for j in range(0, n):
        c2 = crypto.hash_to_scalar(crypto.encodepoint(L1[j]))
        L2 = crypto.point_add(crypto.scalarmult_base(s2[j]), crypto.scalarmult(P2[j], c2))
        LHS = crypto.point_add(LHS, L1[j])
        c1 = crypto.hash_to_scalar(crypto.encodepoint(L2))
        RHS = crypto.point_add(RHS, crypto.scalarmult(P1[j], c1))

    if crypto.point_eq(LHS, RHS):
        return 1
    else:
        logger.warning('Didn\'t verify L1: %s, L1p: %s' % (crypto.encodepoint(LHS), crypto.encodepoint(RHS)))
        return 0




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: https://github.com/monero-project/mininero
# Author: Dusan Klinec, ph4r05, 2018

import logging

from monero_serialize import xmrtypes
from monero_glue.xmr import mlsag2, crypto, common, asnl, monero

logger = logging.getLogger(__name__)
ATOMS = 64

    
def d2b(n, digits):
    b = [0] * digits
    i = 0
    while n:
        b[i] = n & 1
        i = i + 1
        n >>= 1
    return b 


def sum_Ci(Cis):
    """
    Sums points
    :param Cis:
    :return:
    """
    CSum = crypto.identity()
    for i in Cis:
        CSum = crypto.point_add(CSum, i)
    return CSum


def prove_range(amount, last_mask=None, use_asnl=False, mem_opt=True, backend_impl=False, decode=False):
    """
    Range proof generator.
    In order to minimize the memory consumption and CPU usage during transaction generation the returned values
    are returned encoded.

    :param amount:
    :param last_mask:
    :param use_asnl: ASNL range proof, insecure
    :param mem_opt: memory optimized
    :param backend_impl: backend implementation, if available
    :param decode: decodes output
    :return:
    """
    if use_asnl and mem_opt:
        raise ValueError('ASNL not in memory optimized variant')
    if backend_impl and use_asnl:
        raise ValueError('ASNL not supported in backend')

    if backend_impl and crypto.get_backend().has_rangeproof_borromean():
        C, a, R = crypto.prove_range(amount, last_mask)[:3]  # backend returns encoded
        if decode:
            R = monero.recode_rangesig(R, encode=False)
            return C, a, R
        return C, a, R

    ret = None
    if use_asnl and not mem_opt:
        ret = prove_range_orig(amount, last_mask=last_mask, use_asnl=True)
    else:
        ret = prove_range_mem(amount, last_mask=last_mask)

    # Encoding
    C, a, R = ret[:3]
    if not decode:
        R = monero.recode_rangesig(R, encode=True)
    return C, a, R


def prove_range_orig(amount, last_mask=None, use_asnl=False):
    """
    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :param use_asnl: use ASNL, used before Borromean, insecure
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    bb = d2b(amount, ATOMS)  # gives binary form of bb in "digits" binary digits
    logger.info("amount, amount in binary %s %s" % (amount, bb))
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    CiH = [None] * len(bb)  # this is like Ci - 2^i H
    H2 = crypto.gen_Hpow(ATOMS)
    a = crypto.sc_0()
    for i in range(0, ATOMS):
        ai[i] = crypto.random_scalar()
        if last_mask is not None and i == ATOMS - 1:
            ai[i] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(a, ai[i])  # creating the total mask since you have to pass this to receiver...
        if bb[i] == 0:
            Ci[i] = crypto.scalarmult_base(ai[i])
        if bb[i] == 1:
            Ci[i] = crypto.point_add(crypto.scalarmult_base(ai[i]), H2[i])
        CiH[i] = crypto.point_sub(Ci[i], H2[i])

    A = xmrtypes.BoroSig()

    if use_asnl:
        A.s0, A.s1, A.ee = asnl.gen_asnl(ai, Ci, CiH, bb)
    else:
        A.s0, A.s1, A.ee = mlsag2.gen_borromean(ai, Ci, CiH, bb)
    
    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    C = sum_Ci(Ci)
    return C, a, R


def prove_range_mem(amount, last_mask=None):
    """
    Memory optimized range proof.

    Gives C, and mask such that \sumCi = C
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1

    Ci is a commitment to either 0 or 2^i, i=0,...,63
    thus this proves that "amount" is in [0, 2^ATOMS]
    mask is a such that C = aG + bH, and b = amount
    :param amount:
    :param last_mask: ai[ATOMS-1] will be computed as \sum_{i=0}^{ATOMS-2} a_i - last_mask
    :param use_asnl: use ASNL, used before Borromean
    :return: sumCi, mask, RangeSig.
        sumCi is Pedersen commitment on the amount value. sumCi = aG + amount*H
        mask is "a" from the Pedersent commitment above.
    """
    n = ATOMS
    bb = d2b(amount, n)  # gives binary form of bb in "digits" binary digits
    ai = [None] * len(bb)
    Ci = [None] * len(bb)
    a = crypto.sc_0()

    C = crypto.identity()
    alpha = mlsag2.key_zero_vector(n)
    s1 = mlsag2.key_zero_vector(n)
    c_H = crypto.gen_H()
    kck = crypto.get_keccak()  # ee computation

    # First pass, generates: ai, alpha, Ci, ee, s1
    for ii in range(n):
        ai[ii] = crypto.random_scalar()
        if last_mask is not None and ii == ATOMS - 1:
            ai[ii] = crypto.sc_sub(last_mask, a)

        a = crypto.sc_add(a, ai[ii])  # creating the total mask since you have to pass this to receiver...

        alpha[ii] = crypto.random_scalar()
        L = crypto.scalarmult_base(alpha[ii])

        if bb[ii] == 0:
            Ci[ii] = crypto.scalarmult_base(ai[ii])
        else:
            Ci[ii] = crypto.point_add(crypto.scalarmult_base(ai[ii]), c_H)
        C = crypto.point_add(C, Ci[ii])

        if bb[ii] == 0:
            s1[ii] = crypto.random_scalar()
            c = crypto.hash_to_scalar(crypto.encodepoint(L))
            L = crypto.add_keys2(s1[ii], c, crypto.point_sub(Ci[ii], c_H))
            kck.update(crypto.encodepoint(L))

        else:
            kck.update(crypto.encodepoint(L))

        c_H = crypto.point_double(c_H)

    # Compute ee, memory cleanup
    ee = crypto.sc_reduce32(crypto.decodeint(kck.digest()))
    del kck

    # Second phase computes: s0, s1
    c_H = crypto.gen_H()
    s0 = mlsag2.key_zero_vector(n)

    for jj in range(n):
        if not bb[jj]:
            s0[jj] = crypto.sc_mulsub(ai[jj], ee, alpha[jj])

        else:
            s0[jj] = crypto.random_scalar()
            LL = crypto.add_keys2(s0[jj], ee, Ci[jj])
            cc = crypto.hash_to_scalar(crypto.encodepoint(LL))
            s1[jj] = crypto.sc_mulsub(ai[jj], cc, alpha[jj])
        c_H = crypto.point_double(c_H)

    A = xmrtypes.BoroSig()
    A.s0, A.s1, A.ee = s0, s1, ee

    R = xmrtypes.RangeSig()
    R.asig = A
    R.Ci = Ci

    return C, a, R


def ver_range(C=None, rsig=None, use_asnl=False, decode=True):
    """
    Verifies that \sum Ci = C and that each Ci is a commitment to 0 or 2^i
    :param C:
    :param rsig:
    :param use_asnl: use ASNL, used before Borromean, insecure!
    :param decode: decodes encoded range proof
    :return:
    """
    n = ATOMS
    CiH = [None] * n
    C_tmp = crypto.identity()
    c_H = crypto.gen_H()

    if decode:
        rsig = monero.recode_rangesig(rsig, encode=False, copy=True)

    for i in range(0, n):
        CiH[i] = crypto.point_sub(rsig.Ci[i], c_H)
        C_tmp = crypto.point_add(C_tmp, rsig.Ci[i])
        c_H = crypto.point_double(c_H)

    if C is not None and not crypto.point_eq(C_tmp, C):
        return 0

    if use_asnl:
        return asnl.ver_asnl(rsig.Ci, CiH, rsig.asig.s0, rsig.asig.s1, rsig.asig.ee)
    else:
        return mlsag2.ver_borromean(rsig.Ci, CiH, rsig.asig.s0, rsig.asig.s1, rsig.asig.ee)


# Ring-ct MG sigs
# Prove:
#   c.f. http:#eprint.iacr.org/2015/1098 section 4. definition 10. 
#   This does the MG sig on the "dest" part of the given key matrix, and 
#   the last row is the sum of input commitments from that column - sum output commitments
#   this shows that sum inputs = sum outputs
# Ver:
#   verifies the above sig is created corretly


def ecdh_encode(unmasked, receiver_pk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param unmasked:
    :param receiver_pk:
    :param derivation:
    :return:
    """
    rv = xmrtypes.EcdhTuple()
    if derivation is None:
        esk = crypto.random_scalar()
        rv.senderPk = crypto.scalarmult_base(esk)
        derivation = crypto.encodepoint(crypto.scalarmult(receiver_pk, esk))

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_add(unmasked.mask, sharedSec1)
    rv.amount = crypto.sc_add(unmasked.amount, sharedSec2)
    return rv

    
def ecdh_decode(masked, receiver_sk=None, derivation=None):
    """
    Elliptic Curve Diffie-Helman: encodes and decodes the amount b and mask a
    where C= aG + bH
    :param masked:
    :param receiver_sk:
    :param derivation:
    :return:
    """
    rv = xmrtypes.EcdhTuple()

    if derivation is None:
        derivation = crypto.scalarmult(masked.senderPk, receiver_sk)

    sharedSec1 = crypto.hash_to_scalar(derivation)
    sharedSec2 = crypto.hash_to_scalar(crypto.encodeint(sharedSec1))

    rv.mask = crypto.sc_sub(masked.mask, sharedSec1)
    rv.amount = crypto.sc_sub(masked.amount, sharedSec2)
    return rv


#
# RingCT protocol
#


def decode_rct(rv, sk, i):
    """
    c.f. http:#eprint.iacr.org/2015/1098 section 5.1.1
    Uses the attached ecdh info to find the amounts represented by each output commitment
    must know the destination private key to find the correct amount, else will return a random number

    :param rv:
    :param sk:
    :param i:
    :return:
    """
    decodedTuple = ecdh_decode(rv.ecdhInfo[i], sk)
    mask = decodedTuple.mask
    amount = decodedTuple.amount
    C = rv.outPk[i].mask
    H = crypto.gen_H()
    Ctmp = crypto.point_add(crypto.scalarmult_base(mask), crypto.scalarmult(H, amount))
    if not crypto.point_eq(crypto.point_sub(C, Ctmp), crypto.identity()):
        logger.warning("warning, amount decoded incorrectly, will be unable to spend")
    return amount


#
# Key image import / export
#


def generate_ring_signature(prefix_hash, image, pubs, sec, sec_idx, test=False):
    """
    Generates ring signature with key image.
    void crypto_ops::generate_ring_signature()
    TODO: test

    :param prefix_hash:
    :param image:
    :param pubs:
    :param sec:
    :param sec_idx:
    :param test:
    :return:
    """
    if test:
        t = crypto.scalarmult_base(sec)
        if not crypto.point_eq(t, pubs[sec_idx]):
            raise ValueError('Invalid sec key')

        k_i = monero.generate_key_image(crypto.encodepoint(pubs[sec_idx]), sec)
        if not crypto.point_eq(k_i, image):
            raise ValueError('Key image invalid')
        for k in pubs:
            crypto.ge_frombytes_vartime_check(k)

    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff = b'' + prefix_hash
    sum = crypto.sc_0()
    k = crypto.sc_0()
    sig = []
    for i in range(len(pubs)):
        sig.append([crypto.sc_0(), crypto.sc_0()])  # c, r

    for i in range(len(pubs)):
        if i == sec_idx:
            k = crypto.random_scalar()
            tmp3 = crypto.scalarmult_base(k)
            buff += crypto.encodepoint(tmp3)
            tmp3 = crypto.hash_to_ec(crypto.encodepoint(pubs[i]))
            tmp2 = crypto.scalarmult(tmp3, k)
            buff += crypto.encodepoint(tmp2)
        else:
            sig[i] = [crypto.random_scalar(), crypto.random_scalar()]
            tmp3 = crypto.ge_frombytes_vartime(pubs[i])
            tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
            buff += crypto.encodepoint(tmp2)
            tmp3 = crypto.hash_to_ec(crypto.encodepoint(tmp3))
            tmp2 = crypto.ge_double_scalarmult_precomp_vartime(sig[i][1], tmp3, sig[i][0], image_pre)
            buff += crypto.encodepoint(tmp2)
            sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    sig[sec_idx][0] = crypto.sc_sub(h, sum)
    sig[sec_idx][1] = crypto.sc_mulsub(sig[sec_idx][0], sec, k)
    return sig


def check_ring_singature(prefix_hash, image, pubs, sig):
    """
    Checks ring signature generated with generate_ring_signature
    :param prefix_hash:
    :param image:
    :param pubs:
    :param sig:
    :return:
    """
    image_unp = crypto.ge_frombytes_vartime(image)
    image_pre = crypto.ge_dsm_precomp(image_unp)

    buff = b'' + prefix_hash
    sum = crypto.sc_0()
    for i in range(len(pubs)):
        if crypto.sc_check(sig[i][0]) != 0 or crypto.sc_check(sig[i][1]) != 0:
            return False

        tmp3 = crypto.ge_frombytes_vartime(pubs[i])
        tmp2 = crypto.ge_double_scalarmult_base_vartime(sig[i][0], tmp3, sig[i][1])
        buff += crypto.encodepoint(tmp2)
        tmp3 = crypto.hash_to_ec(crypto.encodepoint(pubs[i]))
        tmp2 = crypto.ge_double_scalarmult_precomp_vartime(sig[i][1], tmp3, sig[i][0], image_pre)
        buff += crypto.encodepoint(tmp2)
        sum = crypto.sc_add(sum, sig[i][0])

    h = crypto.hash_to_scalar(buff)
    h = crypto.sc_sub(h, sum)
    return crypto.sc_isnonzero(h) == 0


def export_key_image(creds, subaddresses, pkey, tx_pub_key, additional_tx_pub_keys, out_idx, test=True, verify=True):
    """
    Generates key image for the TXO + signature for the key image
    :param creds:
    :param subaddresses:
    :param pkey:
    :param tx_pub_key:
    :param additional_tx_pub_keys:
    :param out_idx:
    :param test:
    :param verify:
    :return:
    """
    r = monero.generate_key_image_helper(creds, subaddresses, pkey, tx_pub_key, additional_tx_pub_keys, out_idx)
    xi, ki, recv_derivation = r[:3]

    phash = crypto.encodepoint(ki)
    sig = generate_ring_signature(phash, ki, [pkey], xi, 0, test)

    if verify:
        if check_ring_singature(phash, ki, [pkey], sig) != 1:
            raise ValueError('Signature error')

    return ki, sig




#!/usr/bin/env python

import binascii

# list of words from english.h in bitmonero
UNIQUE_PREFIX_LEN = 3
words = [
    'abbey',
    'abducts',
    'ability',
    'ablaze',
    'abnormal',
    'abort',
    'abrasive',
    'absorb',
    'abyss',
    'academy',
    'aces',
    'aching',
    'acidic',
    'acoustic',
    'acquire',
    'across',
    'actress',
    'acumen',
    'adapt',
    'addicted',
    'adept',
    'adhesive',
    'adjust',
    'adopt',
    'adrenalin',
    'adult',
    'adventure',
    'aerial',
    'afar',
    'affair',
    'afield',
    'afloat',
    'afoot',
    'afraid',
    'after',
    'against',
    'agenda',
    'aggravate',
    'agile',
    'aglow',
    'agnostic',
    'agony',
    'agreed',
    'ahead',
    'aided',
    'ailments',
    'aimless',
    'airport',
    'aisle',
    'ajar',
    'akin',
    'alarms',
    'album',
    'alchemy',
    'alerts',
    'algebra',
    'alkaline',
    'alley',
    'almost',
    'aloof',
    'alpine',
    'already',
    'also',
    'altitude',
    'alumni',
    'always',
    'amaze',
    'ambush',
    'amended',
    'amidst',
    'ammo',
    'amnesty',
    'among',
    'amply',
    'amused',
    'anchor',
    'android',
    'anecdote',
    'angled',
    'ankle',
    'annoyed',
    'answers',
    'antics',
    'anvil',
    'anxiety',
    'anybody',
    'apart',
    'apex',
    'aphid',
    'aplomb',
    'apology',
    'apply',
    'apricot',
    'aptitude',
    'aquarium',
    'arbitrary',
    'archer',
    'ardent',
    'arena',
    'argue',
    'arises',
    'army',
    'around',
    'arrow',
    'arsenic',
    'artistic',
    'ascend',
    'ashtray',
    'aside',
    'asked',
    'asleep',
    'aspire',
    'assorted',
    'asylum',
    'athlete',
    'atlas',
    'atom',
    'atrium',
    'attire',
    'auburn',
    'auctions',
    'audio',
    'august',
    'aunt',
    'austere',
    'autumn',
    'avatar',
    'avidly',
    'avoid',
    'awakened',
    'awesome',
    'awful',
    'awkward',
    'awning',
    'awoken',
    'axes',
    'axis',
    'axle',
    'aztec',
    'azure',
    'baby',
    'bacon',
    'badge',
    'baffles',
    'bagpipe',
    'bailed',
    'bakery',
    'balding',
    'bamboo',
    'banjo',
    'baptism',
    'basin',
    'batch',
    'bawled',
    'bays',
    'because',
    'beer',
    'befit',
    'begun',
    'behind',
    'being',
    'below',
    'bemused',
    'benches',
    'berries',
    'bested',
    'betting',
    'bevel',
    'beware',
    'beyond',
    'bias',
    'bicycle',
    'bids',
    'bifocals',
    'biggest',
    'bikini',
    'bimonthly',
    'binocular',
    'biology',
    'biplane',
    'birth',
    'biscuit',
    'bite',
    'biweekly',
    'blender',
    'blip',
    'bluntly',
    'boat',
    'bobsled',
    'bodies',
    'bogeys',
    'boil',
    'boldly',
    'bomb',
    'border',
    'boss',
    'both',
    'bounced',
    'bovine',
    'bowling',
    'boxes',
    'boyfriend',
    'broken',
    'brunt',
    'bubble',
    'buckets',
    'budget',
    'buffet',
    'bugs',
    'building',
    'bulb',
    'bumper',
    'bunch',
    'business',
    'butter',
    'buying',
    'buzzer',
    'bygones',
    'byline',
    'bypass',
    'cabin',
    'cactus',
    'cadets',
    'cafe',
    'cage',
    'cajun',
    'cake',
    'calamity',
    'camp',
    'candy',
    'casket',
    'catch',
    'cause',
    'cavernous',
    'cease',
    'cedar',
    'ceiling',
    'cell',
    'cement',
    'cent',
    'certain',
    'chlorine',
    'chrome',
    'cider',
    'cigar',
    'cinema',
    'circle',
    'cistern',
    'citadel',
    'civilian',
    'claim',
    'click',
    'clue',
    'coal',
    'cobra',
    'cocoa',
    'code',
    'coexist',
    'coffee',
    'cogs',
    'cohesive',
    'coils',
    'colony',
    'comb',
    'cool',
    'copy',
    'corrode',
    'costume',
    'cottage',
    'cousin',
    'cowl',
    'criminal',
    'cube',
    'cucumber',
    'cuddled',
    'cuffs',
    'cuisine',
    'cunning',
    'cupcake',
    'custom',
    'cycling',
    'cylinder',
    'cynical',
    'dabbing',
    'dads',
    'daft',
    'dagger',
    'daily',
    'damp',
    'dangerous',
    'dapper',
    'darted',
    'dash',
    'dating',
    'dauntless',
    'dawn',
    'daytime',
    'dazed',
    'debut',
    'decay',
    'dedicated',
    'deepest',
    'deftly',
    'degrees',
    'dehydrate',
    'deity',
    'dejected',
    'delayed',
    'demonstrate',
    'dented',
    'deodorant',
    'depth',
    'desk',
    'devoid',
    'dewdrop',
    'dexterity',
    'dialect',
    'dice',
    'diet',
    'different',
    'digit',
    'dilute',
    'dime',
    'dinner',
    'diode',
    'diplomat',
    'directed',
    'distance',
    'ditch',
    'divers',
    'dizzy',
    'doctor',
    'dodge',
    'does',
    'dogs',
    'doing',
    'dolphin',
    'domestic',
    'donuts',
    'doorway',
    'dormant',
    'dosage',
    'dotted',
    'double',
    'dove',
    'down',
    'dozen',
    'dreams',
    'drinks',
    'drowning',
    'drunk',
    'drying',
    'dual',
    'dubbed',
    'duckling',
    'dude',
    'duets',
    'duke',
    'dullness',
    'dummy',
    'dunes',
    'duplex',
    'duration',
    'dusted',
    'duties',
    'dwarf',
    'dwelt',
    'dwindling',
    'dying',
    'dynamite',
    'dyslexic',
    'each',
    'eagle',
    'earth',
    'easy',
    'eating',
    'eavesdrop',
    'eccentric',
    'echo',
    'eclipse',
    'economics',
    'ecstatic',
    'eden',
    'edgy',
    'edited',
    'educated',
    'eels',
    'efficient',
    'eggs',
    'egotistic',
    'eight',
    'either',
    'eject',
    'elapse',
    'elbow',
    'eldest',
    'eleven',
    'elite',
    'elope',
    'else',
    'eluded',
    'emails',
    'ember',
    'emerge',
    'emit',
    'emotion',
    'empty',
    'emulate',
    'energy',
    'enforce',
    'enhanced',
    'enigma',
    'enjoy',
    'enlist',
    'enmity',
    'enough',
    'enraged',
    'ensign',
    'entrance',
    'envy',
    'epoxy',
    'equip',
    'erase',
    'erected',
    'erosion',
    'error',
    'eskimos',
    'espionage',
    'essential',
    'estate',
    'etched',
    'eternal',
    'ethics',
    'etiquette',
    'evaluate',
    'evenings',
    'evicted',
    'evolved',
    'examine',
    'excess',
    'exhale',
    'exit',
    'exotic',
    'exquisite',
    'extra',
    'exult',
    'fabrics',
    'factual',
    'fading',
    'fainted',
    'faked',
    'fall',
    'family',
    'fancy',
    'farming',
    'fatal',
    'faulty',
    'fawns',
    'faxed',
    'fazed',
    'feast',
    'february',
    'federal',
    'feel',
    'feline',
    'females',
    'fences',
    'ferry',
    'festival',
    'fetches',
    'fever',
    'fewest',
    'fiat',
    'fibula',
    'fictional',
    'fidget',
    'fierce',
    'fifteen',
    'fight',
    'films',
    'firm',
    'fishing',
    'fitting',
    'five',
    'fixate',
    'fizzle',
    'fleet',
    'flippant',
    'flying',
    'foamy',
    'focus',
    'foes',
    'foggy',
    'foiled',
    'folding',
    'fonts',
    'foolish',
    'fossil',
    'fountain',
    'fowls',
    'foxes',
    'foyer',
    'framed',
    'friendly',
    'frown',
    'fruit',
    'frying',
    'fudge',
    'fuel',
    'fugitive',
    'fully',
    'fuming',
    'fungal',
    'furnished',
    'fuselage',
    'future',
    'fuzzy',
    'gables',
    'gadget',
    'gags',
    'gained',
    'galaxy',
    'gambit',
    'gang',
    'gasp',
    'gather',
    'gauze',
    'gave',
    'gawk',
    'gaze',
    'gearbox',
    'gecko',
    'geek',
    'gels',
    'gemstone',
    'general',
    'geometry',
    'germs',
    'gesture',
    'getting',
    'geyser',
    'ghetto',
    'ghost',
    'giant',
    'giddy',
    'gifts',
    'gigantic',
    'gills',
    'gimmick',
    'ginger',
    'girth',
    'giving',
    'glass',
    'gleeful',
    'glide',
    'gnaw',
    'gnome',
    'goat',
    'goblet',
    'godfather',
    'goes',
    'goggles',
    'going',
    'goldfish',
    'gone',
    'goodbye',
    'gopher',
    'gorilla',
    'gossip',
    'gotten',
    'gourmet',
    'governing',
    'gown',
    'greater',
    'grunt',
    'guarded',
    'guest',
    'guide',
    'gulp',
    'gumball',
    'guru',
    'gusts',
    'gutter',
    'guys',
    'gymnast',
    'gypsy',
    'gyrate',
    'habitat',
    'hacksaw',
    'haggled',
    'hairy',
    'hamburger',
    'happens',
    'hashing',
    'hatchet',
    'haunted',
    'having',
    'hawk',
    'haystack',
    'hazard',
    'hectare',
    'hedgehog',
    'heels',
    'hefty',
    'height',
    'hemlock',
    'hence',
    'heron',
    'hesitate',
    'hexagon',
    'hickory',
    'hiding',
    'highway',
    'hijack',
    'hiker',
    'hills',
    'himself',
    'hinder',
    'hippo',
    'hire',
    'history',
    'hitched',
    'hive',
    'hoax',
    'hobby',
    'hockey',
    'hoisting',
    'hold',
    'honked',
    'hookup',
    'hope',
    'hornet',
    'hospital',
    'hotel',
    'hounded',
    'hover',
    'howls',
    'hubcaps',
    'huddle',
    'huge',
    'hull',
    'humid',
    'hunter',
    'hurried',
    'husband',
    'huts',
    'hybrid',
    'hydrogen',
    'hyper',
    'iceberg',
    'icing',
    'icon',
    'identity',
    'idiom',
    'idled',
    'idols',
    'igloo',
    'ignore',
    'iguana',
    'illness',
    'imagine',
    'imbalance',
    'imitate',
    'impel',
    'inactive',
    'inbound',
    'incur',
    'industrial',
    'inexact',
    'inflamed',
    'ingested',
    'initiate',
    'injury',
    'inkling',
    'inline',
    'inmate',
    'innocent',
    'inorganic',
    'input',
    'inquest',
    'inroads',
    'insult',
    'intended',
    'inundate',
    'invoke',
    'inwardly',
    'ionic',
    'irate',
    'iris',
    'irony',
    'irritate',
    'island',
    'isolated',
    'issued',
    'italics',
    'itches',
    'items',
    'itinerary',
    'itself',
    'ivory',
    'jabbed',
    'jackets',
    'jaded',
    'jagged',
    'jailed',
    'jamming',
    'january',
    'jargon',
    'jaunt',
    'javelin',
    'jaws',
    'jazz',
    'jeans',
    'jeers',
    'jellyfish',
    'jeopardy',
    'jerseys',
    'jester',
    'jetting',
    'jewels',
    'jigsaw',
    'jingle',
    'jittery',
    'jive',
    'jobs',
    'jockey',
    'jogger',
    'joining',
    'joking',
    'jolted',
    'jostle',
    'journal',
    'joyous',
    'jubilee',
    'judge',
    'juggled',
    'juicy',
    'jukebox',
    'july',
    'jump',
    'junk',
    'jury',
    'justice',
    'juvenile',
    'kangaroo',
    'karate',
    'keep',
    'kennel',
    'kept',
    'kernels',
    'kettle',
    'keyboard',
    'kickoff',
    'kidneys',
    'king',
    'kiosk',
    'kisses',
    'kitchens',
    'kiwi',
    'knapsack',
    'knee',
    'knife',
    'knowledge',
    'knuckle',
    'koala',
    'laboratory',
    'ladder',
    'lagoon',
    'lair',
    'lakes',
    'lamb',
    'language',
    'laptop',
    'large',
    'last',
    'later',
    'launching',
    'lava',
    'lawsuit',
    'layout',
    'lazy',
    'lectures',
    'ledge',
    'leech',
    'left',
    'legion',
    'leisure',
    'lemon',
    'lending',
    'leopard',
    'lesson',
    'lettuce',
    'lexicon',
    'liar',
    'library',
    'licks',
    'lids',
    'lied',
    'lifestyle',
    'light',
    'likewise',
    'lilac',
    'limits',
    'linen',
    'lion',
    'lipstick',
    'liquid',
    'listen',
    'lively',
    'loaded',
    'lobster',
    'locker',
    'lodge',
    'lofty',
    'logic',
    'loincloth',
    'long',
    'looking',
    'lopped',
    'lordship',
    'losing',
    'lottery',
    'loudly',
    'love',
    'lower',
    'loyal',
    'lucky',
    'luggage',
    'lukewarm',
    'lullaby',
    'lumber',
    'lunar',
    'lurk',
    'lush',
    'luxury',
    'lymph',
    'lynx',
    'lyrics',
    'macro',
    'madness',
    'magically',
    'mailed',
    'major',
    'makeup',
    'malady',
    'mammal',
    'maps',
    'masterful',
    'match',
    'maul',
    'maverick',
    'maximum',
    'mayor',
    'maze',
    'meant',
    'mechanic',
    'medicate',
    'meeting',
    'megabyte',
    'melting',
    'memoir',
    'menu',
    'merger',
    'mesh',
    'metro',
    'mews',
    'mice',
    'midst',
    'mighty',
    'mime',
    'mirror',
    'misery',
    'mittens',
    'mixture',
    'moat',
    'mobile',
    'mocked',
    'mohawk',
    'moisture',
    'molten',
    'moment',
    'money',
    'moon',
    'mops',
    'morsel',
    'mostly',
    'motherly',
    'mouth',
    'movement',
    'mowing',
    'much',
    'muddy',
    'muffin',
    'mugged',
    'mullet',
    'mumble',
    'mundane',
    'muppet',
    'mural',
    'musical',
    'muzzle',
    'myriad',
    'mystery',
    'myth',
    'nabbing',
    'nagged',
    'nail',
    'names',
    'nanny',
    'napkin',
    'narrate',
    'nasty',
    'natural',
    'nautical',
    'navy',
    'nearby',
    'necklace',
    'needed',
    'negative',
    'neither',
    'neon',
    'nephew',
    'nerves',
    'nestle',
    'network',
    'neutral',
    'never',
    'newt',
    'nexus',
    'nibs',
    'niche',
    'niece',
    'nifty',
    'nightly',
    'nimbly',
    'nineteen',
    'nirvana',
    'nitrogen',
    'nobody',
    'nocturnal',
    'nodes',
    'noises',
    'nomad',
    'noodles',
    'northern',
    'nostril',
    'noted',
    'nouns',
    'novelty',
    'nowhere',
    'nozzle',
    'nuance',
    'nucleus',
    'nudged',
    'nugget',
    'nuisance',
    'null',
    'number',
    'nuns',
    'nurse',
    'nutshell',
    'nylon',
    'oaks',
    'oars',
    'oasis',
    'oatmeal',
    'obedient',
    'object',
    'obliged',
    'obnoxious',
    'observant',
    'obtains',
    'obvious',
    'occur',
    'ocean',
    'october',
    'odds',
    'odometer',
    'offend',
    'often',
    'oilfield',
    'ointment',
    'okay',
    'older',
    'olive',
    'olympics',
    'omega',
    'omission',
    'omnibus',
    'onboard',
    'oncoming',
    'oneself',
    'ongoing',
    'onion',
    'online',
    'onslaught',
    'onto',
    'onward',
    'oozed',
    'opacity',
    'opened',
    'opposite',
    'optical',
    'opus',
    'orange',
    'orbit',
    'orchid',
    'orders',
    'organs',
    'origin',
    'ornament',
    'orphans',
    'oscar',
    'ostrich',
    'otherwise',
    'otter',
    'ouch',
    'ought',
    'ounce',
    'ourselves',
    'oust',
    'outbreak',
    'oval',
    'oven',
    'owed',
    'owls',
    'owner',
    'oxidant',
    'oxygen',
    'oyster',
    'ozone',
    'pact',
    'paddles',
    'pager',
    'pairing',
    'palace',
    'pamphlet',
    'pancakes',
    'paper',
    'paradise',
    'pastry',
    'patio',
    'pause',
    'pavements',
    'pawnshop',
    'payment',
    'peaches',
    'pebbles',
    'peculiar',
    'pedantic',
    'peeled',
    'pegs',
    'pelican',
    'pencil',
    'people',
    'pepper',
    'perfect',
    'pests',
    'petals',
    'phase',
    'pheasants',
    'phone',
    'phrases',
    'physics',
    'piano',
    'picked',
    'pierce',
    'pigment',
    'piloted',
    'pimple',
    'pinched',
    'pioneer',
    'pipeline',
    'pirate',
    'pistons',
    'pitched',
    'pivot',
    'pixels',
    'pizza',
    'playful',
    'pledge',
    'pliers',
    'plotting',
    'plus',
    'plywood',
    'poaching',
    'pockets',
    'podcast',
    'poetry',
    'point',
    'poker',
    'polar',
    'ponies',
    'pool',
    'popular',
    'portents',
    'possible',
    'potato',
    'pouch',
    'poverty',
    'powder',
    'pram',
    'present',
    'pride',
    'problems',
    'pruned',
    'prying',
    'psychic',
    'public',
    'puck',
    'puddle',
    'puffin',
    'pulp',
    'pumpkins',
    'punch',
    'puppy',
    'purged',
    'push',
    'putty',
    'puzzled',
    'pylons',
    'pyramid',
    'python',
    'queen',
    'quick',
    'quote',
    'rabbits',
    'racetrack',
    'radar',
    'rafts',
    'rage',
    'railway',
    'raking',
    'rally',
    'ramped',
    'randomly',
    'rapid',
    'rarest',
    'rash',
    'rated',
    'ravine',
    'rays',
    'razor',
    'react',
    'rebel',
    'recipe',
    'reduce',
    'reef',
    'refer',
    'regular',
    'reheat',
    'reinvest',
    'rejoices',
    'rekindle',
    'relic',
    'remedy',
    'renting',
    'reorder',
    'repent',
    'request',
    'reruns',
    'rest',
    'return',
    'reunion',
    'revamp',
    'rewind',
    'rhino',
    'rhythm',
    'ribbon',
    'richly',
    'ridges',
    'rift',
    'rigid',
    'rims',
    'ringing',
    'riots',
    'ripped',
    'rising',
    'ritual',
    'river',
    'roared',
    'robot',
    'rockets',
    'rodent',
    'rogue',
    'roles',
    'romance',
    'roomy',
    'roped',
    'roster',
    'rotate',
    'rounded',
    'rover',
    'rowboat',
    'royal',
    'ruby',
    'rudely',
    'ruffled',
    'rugged',
    'ruined',
    'ruling',
    'rumble',
    'runway',
    'rural',
    'rustled',
    'ruthless',
    'sabotage',
    'sack',
    'sadness',
    'safety',
    'saga',
    'sailor',
    'sake',
    'salads',
    'sample',
    'sanity',
    'sapling',
    'sarcasm',
    'sash',
    'satin',
    'saucepan',
    'saved',
    'sawmill',
    'saxophone',
    'sayings',
    'scamper',
    'scenic',
    'school',
    'science',
    'scoop',
    'scrub',
    'scuba',
    'seasons',
    'second',
    'sedan',
    'seeded',
    'segments',
    'seismic',
    'selfish',
    'semifinal',
    'sensible',
    'september',
    'sequence',
    'serving',
    'session',
    'setup',
    'seventh',
    'sewage',
    'shackles',
    'shelter',
    'shipped',
    'shocking',
    'shrugged',
    'shuffled',
    'shyness',
    'siblings',
    'sickness',
    'sidekick',
    'sieve',
    'sifting',
    'sighting',
    'silk',
    'simplest',
    'sincerely',
    'sipped',
    'siren',
    'situated',
    'sixteen',
    'sizes',
    'skater',
    'skew',
    'skirting',
    'skulls',
    'skydive',
    'slackens',
    'sleepless',
    'slid',
    'slower',
    'slug',
    'smash',
    'smelting',
    'smidgen',
    'smog',
    'smuggled',
    'snake',
    'sneeze',
    'sniff',
    'snout',
    'snug',
    'soapy',
    'sober',
    'soccer',
    'soda',
    'software',
    'soggy',
    'soil',
    'solved',
    'somewhere',
    'sonic',
    'soothe',
    'soprano',
    'sorry',
    'southern',
    'sovereign',
    'sowed',
    'soya',
    'space',
    'speedy',
    'sphere',
    'spiders',
    'splendid',
    'spout',
    'sprig',
    'spud',
    'spying',
    'square',
    'stacking',
    'stellar',
    'stick',
    'stockpile',
    'strained',
    'stunning',
    'stylishly',
    'subtly',
    'succeed',
    'suddenly',
    'suede',
    'suffice',
    'sugar',
    'suitcase',
    'sulking',
    'summon',
    'sunken',
    'superior',
    'surfer',
    'sushi',
    'suture',
    'swagger',
    'swept',
    'swiftly',
    'sword',
    'swung',
    'syllabus',
    'symptoms',
    'syndrome',
    'syringe',
    'system',
    'taboo',
    'tacit',
    'tadpoles',
    'tagged',
    'tail',
    'taken',
    'talent',
    'tamper',
    'tanks',
    'tapestry',
    'tarnished',
    'tasked',
    'tattoo',
    'taunts',
    'tavern',
    'tawny',
    'taxi',
    'teardrop',
    'technical',
    'tedious',
    'teeming',
    'tell',
    'template',
    'tender',
    'tepid',
    'tequila',
    'terminal',
    'testing',
    'tether',
    'textbook',
    'thaw',
    'theatrics',
    'thirsty',
    'thorn',
    'threaten',
    'thumbs',
    'thwart',
    'ticket',
    'tidy',
    'tiers',
    'tiger',
    'tilt',
    'timber',
    'tinted',
    'tipsy',
    'tirade',
    'tissue',
    'titans',
    'toaster',
    'tobacco',
    'today',
    'toenail',
    'toffee',
    'together',
    'toilet',
    'token',
    'tolerant',
    'tomorrow',
    'tonic',
    'toolbox',
    'topic',
    'torch',
    'tossed',
    'total',
    'touchy',
    'towel',
    'toxic',
    'toyed',
    'trash',
    'trendy',
    'tribal',
    'trolling',
    'truth',
    'trying',
    'tsunami',
    'tubes',
    'tucks',
    'tudor',
    'tuesday',
    'tufts',
    'tugs',
    'tuition',
    'tulips',
    'tumbling',
    'tunnel',
    'turnip',
    'tusks',
    'tutor',
    'tuxedo',
    'twang',
    'tweezers',
    'twice',
    'twofold',
    'tycoon',
    'typist',
    'tyrant',
    'ugly',
    'ulcers',
    'ultimate',
    'umbrella',
    'umpire',
    'unafraid',
    'unbending',
    'uncle',
    'under',
    'uneven',
    'unfit',
    'ungainly',
    'unhappy',
    'union',
    'unjustly',
    'unknown',
    'unlikely',
    'unmask',
    'unnoticed',
    'unopened',
    'unplugs',
    'unquoted',
    'unrest',
    'unsafe',
    'until',
    'unusual',
    'unveil',
    'unwind',
    'unzip',
    'upbeat',
    'upcoming',
    'update',
    'upgrade',
    'uphill',
    'upkeep',
    'upload',
    'upon',
    'upper',
    'upright',
    'upstairs',
    'uptight',
    'upwards',
    'urban',
    'urchins',
    'urgent',
    'usage',
    'useful',
    'usher',
    'using',
    'usual',
    'utensils',
    'utility',
    'utmost',
    'utopia',
    'uttered',
    'vacation',
    'vague',
    'vain',
    'value',
    'vampire',
    'vane',
    'vapidly',
    'vary',
    'vastness',
    'vats',
    'vaults',
    'vector',
    'veered',
    'vegan',
    'vehicle',
    'vein',
    'velvet',
    'venomous',
    'verification',
    'vessel',
    'veteran',
    'vexed',
    'vials',
    'vibrate',
    'victim',
    'video',
    'viewpoint',
    'vigilant',
    'viking',
    'village',
    'vinegar',
    'violin',
    'vipers',
    'virtual',
    'visited',
    'vitals',
    'vivid',
    'vixen',
    'vocal',
    'vogue',
    'voice',
    'volcano',
    'vortex',
    'voted',
    'voucher',
    'vowels',
    'voyage',
    'vulture',
    'wade',
    'waffle',
    'wagtail',
    'waist',
    'waking',
    'wallets',
    'wanted',
    'warped',
    'washing',
    'water',
    'waveform',
    'waxing',
    'wayside',
    'weavers',
    'website',
    'wedge',
    'weekday',
    'weird',
    'welders',
    'went',
    'wept',
    'were',
    'western',
    'wetsuit',
    'whale',
    'when',
    'whipped',
    'whole',
    'wickets',
    'width',
    'wield',
    'wife',
    'wiggle',
    'wildly',
    'winter',
    'wipeout',
    'wiring',
    'wise',
    'withdrawn',
    'wives',
    'wizard',
    'wobbly',
    'woes',
    'woken',
    'wolf',
    'womanly',
    'wonders',
    'woozy',
    'worry',
    'wounded',
    'woven',
    'wrap',
    'wrist',
    'wrong',
    'yacht',
    'yahoo',
    'yanks',
    'yard',
    'yawning',
    'yearbook',
    'yellow',
    'yesterday',
    'yeti',
    'yields',
    'yodel',
    'yoga',
    'younger',
    'yoyo',
    'zapped',
    'zeal',
    'zebra',
    'zero',
    'zesty',
    'zigzags',
    'zinger',
    'zippers',
    'zodiac',
    'zombie',
    'zones',
    'zoom'
]


n = len(words)


def mn_len():
    return len(words)


def mn_words():
    return words


# Note about US patent no 5892470: Here each word does not represent a given digit.
# Instead, the digit represented by a word is variable, it depends on the previous word.


def mn_swap_endian_4byte(st):
    # this is from moneromoo's code
    r = st[6:8]+st[4:6]+st[2:4]+st[0:2]
    return r


def mn_encode(message, checksum=False):
    out = []
    n = len(words)
    message = binascii.hexlify(message)
    for i in range(0, len(message), 8):
        message = message[0:i] + mn_swap_endian_4byte(message[i:i+8]) + message[i+8:]

    for i in range(len(message)//8):
        word = message[8*i:8*i+8]
        x = int(word, 16)
        w1 = x % n
        w2 = ((x//n) + w1) % n
        w3 = (((x//n)//n) + w2) % n
        out += [words[w1], words[w2], words[w3]]

    if checksum:
        idx = checksum_create(out)
        out.append(out[idx])
    return out


def mn_decode(wlist):
    out = ''
    n = len(words)
    has_checksum = (n % 3) == 1

    if has_checksum:
        if not checksum_test(wlist):
            raise ValueError('Electrum checksum incorrect')
        wlist = wlist[:-1]

    for i in range(len(wlist)//3):  # note 24 / 3 = 8... 12 / 3 = 4..
        word1, word2, word3 = wlist[3*i:3*i+3]
        w1 = words.index(word1)
        w2 = words.index(word2)
        w3 = words.index(word3)
        x = w1 + n * ((n + w2 - w1) % n) +n * n * ((n + w3 - w2) % n)  # as an int
        b = '%08x' % x  # this is big endian!
        out += mn_swap_endian_4byte(b)

    return out


def checksum_create(wlist, unique_prefix_len=UNIQUE_PREFIX_LEN):
    acc = ''
    for w in wlist:
        acc += w[:unique_prefix_len] if len(w) > unique_prefix_len else w

    acc = acc.encode('utf8')
    crc = binascii.crc32(acc) & 0xFFFFFFFF
    return crc % len(wlist)


def checksum_test(wlist, unique_prefix_len=UNIQUE_PREFIX_LEN):
    words = wlist
    last_word = words.pop()
    checksum = words[checksum_create(words, unique_prefix_len)]
    trimmed_checksum = checksum[:unique_prefix_len] if len(checksum) > unique_prefix_len else checksum
    trimmed_last_word = last_word[:unique_prefix_len] if len(last_word) > unique_prefix_len else last_word
    return trimmed_checksum == trimmed_last_word






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018


import collections


class CallTracker(object):
    """
    Tracks method invocations in the given object / module
    """
    def __init__(self, base, print_on_end=False):
        self.base = base
        self.t = collections.defaultdict(lambda: 0)
        self.print_on_end = print_on_end

    def __getattr__(self, name):
        def handler(*args, **kwargs):
            self.t[name] += 1
            return getattr(self.base, name)(*args, **kwargs)

        return handler

    def __del__(self):
        if self.print_on_end:
            k = sorted(self.t.keys())
            for kk in k:
                print('  %s: %s' % (kk, self.t[kk]))

    def call_tracker_get_call_stats(self):
        return self.t





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest
import aiounittest

from monero_glue.xmr import crypto
from monero_glue.xmr.enc import aesgcm


class AesTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(AesTest, self).__init__(*args, **kwargs)

    def test_enc(self):
        key = crypto.keccak_hash(b'0')
        plain = b'1234'
        assoc = b'4567'
        iv, cip, tag = aesgcm.encrypt(key, plain, assoc)

        plain2 = aesgcm.decrypt(key, iv, cip, tag, assoc)
        self.assertEqual(plain, plain2)

        with self.assertRaises(ValueError):
            plain2 = aesgcm.decrypt(key, iv, cip, tag, assoc + b'1')

        with self.assertRaises(ValueError):
            plain2 = aesgcm.decrypt(key, bytes(iv[0] ^ 0xff) + iv[1:], cip, tag, assoc)

        with self.assertRaises(ValueError):
            plain2 = aesgcm.decrypt(key, iv, bytes(cip[0] ^ 0xff) + cip[1:], tag, assoc)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





from binascii import hexlify
from binascii import unhexlify
from hashlib import sha256
from hashlib import sha512
import hmac

from monero_glue.xmr import crypto
from monero_glue.misc import b58 as base58
from monero_glue.misc.bip import bip39
from os import urandom
from ecdsa import SECP256k1
from ecdsa.ecdsa import Public_key as _ECDSA_Public_key
from ecdsa.ellipticcurve import INFINITY
import six
import time

# import all the networks
from .network import *

from .keys import incompatible_network_exception_factory
from .keys import PrivateKey
from .keys import PublicKey
from .keys import PublicPair
from .utils import chr_py2
from .utils import ensure_bytes
from .utils import ensure_str
from .utils import hash160
from .utils import is_hex_string
from .utils import long_or_int
from .utils import long_to_hex
from .utils import memoize


class Wallet(object):
    """A BIP32 wallet is made up of Wallet nodes.
    A Private node contains both a public and private key, while a public
    node contains only a public key.
    **WARNING**:
    When creating a NEW wallet you MUST back up the private key. If
    you don't then any coins sent to your address will be LOST FOREVER.
    You need to save the private key somewhere. It is OK to just write
    it down on a piece of paper! Don't share this key with anyone!
    >>> my_wallet = Wallet.from_master_secret(
    ...     key='correct horse battery staple')
    >>> private = my_wallet.serialize(private=True)
    >>> private  # doctest: +ELLIPSIS
    u'xprv9s21ZrQH143K2mDJW8vDeFwbyDbFv868mM2Zr87rJSTj8q16Unkaq1pryiV...'
    If you want to use this wallet on your website to accept bitcoin or
    altcoin payments, you should first create a primary child.
    BIP32 Hierarchical Deterministic Wallets are described in this BIP:
    https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki
    """

    def __init__(self,
                 chain_code,
                 depth=0,
                 parent_fingerprint=0,
                 child_number=0,
                 private_exponent=None,
                 private_key=None,
                 public_pair=None,
                 public_key=None,
                 network="bitcoin_testnet",
                 seed_secret=None):
        """Construct a new BIP32 compliant wallet.
        You probably don't want to use this init methd. Instead use one
        of the 'from_master_secret' or 'deserialize' cosntructors.
        """

        if (not (private_exponent or private_key) and
                not (public_pair or public_key)):
            raise InsufficientKeyDataError(
                "You must supply one of private_exponent or public_pair")

        network = Wallet.get_network(network)
        self.private_key = None
        self.public_key = None
        self.seed_secret = seed_secret
        if private_key:
            if not isinstance(private_key, PrivateKey):
                raise InvalidPrivateKeyError(
                    "private_key must be of type "
                    "bitmerchant.wallet.keys.PrivateKey")
            self.private_key = private_key
        elif private_exponent:
            self.private_key = PrivateKey(
                private_exponent, network=network)

        if public_key:
            if not isinstance(public_key, PublicKey):
                raise InvalidPublicKeyError(
                    "public_key must be of type "
                    "bitmerchant.wallet.keys.PublicKey")
            self.public_key = public_key
        elif public_pair:
            self.public_key = PublicKey.from_public_pair(
                public_pair, network=network)
        else:
            self.public_key = self.private_key.get_public_key()

        if (self.private_key and self.private_key.get_public_key() !=
                self.public_key):
            raise KeyMismatchError(
                "Provided private and public values do not match")

        def h(val, hex_len):
            if isinstance(val, six.integer_types):
                return long_to_hex(val, hex_len)
            elif (isinstance(val, six.string_types) or
                    isinstance(val, six.binary_type)) and is_hex_string(val):
                val = ensure_bytes(val)
                if len(val) != hex_len:
                    raise ValueError("Invalid parameter length")
                return val
            else:
                raise ValueError("Invalid parameter type")

        def l(val):
            if isinstance(val, six.integer_types):
                return long_or_int(val)
            elif (isinstance(val, six.string_types) or
                    isinstance(val, six.binary_type)):
                val = ensure_bytes(val)
                if not is_hex_string(val):
                    val = hexlify(val)
                return long_or_int(val, 16)
            else:
                raise ValueError("parameter must be an int or long")

        self.network = Wallet.get_network(network)
        self.depth = l(depth)
        if (isinstance(parent_fingerprint, six.string_types) or
                isinstance(parent_fingerprint, six.binary_type)):
            val = ensure_bytes(parent_fingerprint)
            if val.startswith(b"0x"):
                parent_fingerprint = val[2:]
        self.parent_fingerprint = b"0x" + h(parent_fingerprint, 8)
        self.child_number = l(child_number)
        self.chain_code = h(chain_code, 64)

    def get_private_key_hex(self):
        """
        Get the hex-encoded (I guess SEC1?) representation of the private key.
        DO NOT share this private key with anyone.
        """
        return ensure_bytes(self.private_key.get_key())

    def get_public_key_hex(self, compressed=True):
        """Get the sec1 representation of the public key."""
        return ensure_bytes(self.public_key.get_key(compressed))

    @property
    def identifier(self):
        """Get the identifier for this node.
        Extended keys can be identified by the Hash160 (RIPEMD160 after SHA256)
        of the public key's `key`. This corresponds exactly to the data used in
        traditional Bitcoin addresses. It is not advised to represent this data
        in base58 format though, as it may be interpreted as an address that
        way (and wallet software is not required to accept payment to the chain
        key itself).
        """
        key = self.get_public_key_hex()
        return ensure_bytes(hexlify(hash160(unhexlify(key))))

    @property
    def fingerprint(self):
        """The first 32 bits of the identifier are called the fingerprint."""
        # 32 bits == 4 Bytes == 8 hex characters
        return b'0x' + self.identifier[:8]

    @staticmethod
    def bytes_to_indices(seed, bits=11):
        """
        Transforms byte stream to list of bits-sized integers.
        Used to transform secret seed to word list.
        Consistent with https://iancoleman.io/bip39/
        
        :param seed:
        :param bits:
        :return:
        """
        seed_len = len(seed)
        indices = []
        carry = 0
        carry_bits = 0
        for idx, b in enumerate(seed):
            cleft = min(8, bits - carry_bits)
            cleft_bl = 8 - cleft
            flush = cleft == bits - carry_bits

            carry_bits += cleft
            carry <<= cleft
            carry |= (int(b) >> cleft_bl) & (2 ** cleft - 1)

            if idx + 1 == seed_len and cleft != 0:
                carry <<= bits - carry_bits
                flush = True

            if flush:
                indices.append(carry)
                carry = int(b) & (2 ** cleft_bl - 1)
                carry_bits = cleft_bl

        return indices

    @staticmethod
    def indices_to_bytes(indices, bits=11):
        """
        Converts indices to byte array
        :param indices:
        :param bits:
        :return:
        """
        ind_len = len(indices)
        barr = []

        cidx = 0
        carry = 0
        carry_bits = 0
        avail = bits

        while True:
            cb = 8 - carry_bits
            cons = min(cb, avail)
            carry <<= cons
            carry |= indices[cidx] >> (avail - cons) & ((1 << cons) - 1)
            carry_bits += cons
            avail -= cons

            if cb == cons:
                barr.append(carry)
                carry = 0
                carry_bits = 0

            if avail == 0:
                if cidx + 1 == ind_len:
                    break
                cidx += 1
                avail = bits
        return bytes(barr)

    def to_seed_words(self):
        """
        Generates word-based seed:
        secret seed bits + SHA256 checksum -> words.
        :return:
        """
        if self.seed_secret is None:
            return None

        checksum = sha256(self.seed_secret).digest()
        seed = self.seed_secret + checksum[:1]

        indices = Wallet.bytes_to_indices(seed)
        words = [bip39.english_words[x] for x in indices]
        return words, indices

    def create_new_address_for_user(self, user_id):
        """Create a new bitcoin address to accept payments for a User.
        This is a convenience wrapper around `get_child` that helps you do
        the right thing. This method always creates a public, non-prime
        address that can be generated from a BIP32 public key on an
        insecure server."""
        max_id = 0x80000000
        if user_id < 0 or user_id > max_id:
            raise ValueError(
                "Invalid UserID. Must be between 0 and %s" % max_id)
        return self.get_child(user_id, is_prime=False, as_private=False)

    def get_child_for_path(self, path):
        """Get a child for a given path.
        Rather than repeated calls to get_child, children can be found
        by a derivation path. Paths look like:
            m/0/1'/10
        Which is the same as
            self.get_child(0).get_child(-1).get_child(10)
        Or, in other words, the 10th publicly derived child of the 1st
        privately derived child of the 0th publicly derived child of master.
        You can use either ' or p to denote a prime (that is, privately
        derived) child.
        A child that has had its private key stripped can be requested by
        either passing a capital M or appending '.pub' to the end of the path.
        These three paths all give the same child that has had its private
        key scrubbed:
            M/0/1
            m/0/1.pub
            M/0/1.pub
        """
        path = ensure_str(path)

        if not path:
            raise InvalidPathError("%s is not a valid path" % path)

        # Figure out public/private derivation
        as_private = True
        if path.startswith("M"):
            as_private = False
        if path.endswith(".pub"):
            as_private = False
            path = path[:-4]

        parts = path.split("/")
        if len(parts) == 0:
            raise InvalidPathError()

        child = self
        for part in parts:
            if part.lower() == "m":
                continue
            is_prime = None  # Let primeness be figured out by the child number
            if part[-1] in "'p":
                is_prime = True
                part = part.replace("'", "").replace("p", "")
            try:
                child_number = long_or_int(part)
            except ValueError:
                raise InvalidPathError("%s is not a valid path" % path)
            child = child.get_child(child_number, is_prime)
        if not as_private:
            return child.public_copy()
        return child

    @memoize
    def get_child(self, child_number, is_prime=None, as_private=True):
        """Derive a child key.
        :param child_number: The number of the child key to compute
        :type child_number: int
        :param is_prime: If True, the child is calculated via private
            derivation. If False, then public derivation is used. If None,
            then it is figured out from the value of child_number.
        :type is_prime: bool, defaults to None
        :param as_private: If True, strips private key from the result.
            Defaults to False. If there is no private key present, this is
            ignored.
        :type as_private: bool
        Positive child_numbers (less than 2,147,483,648) produce publicly
        derived children.
        Negative numbers (greater than -2,147,483,648) uses private derivation.
        NOTE: Python can't do -0, so if you want the privately derived 0th
        child you need to manually set is_prime=True.
        NOTE: negative numbered children are provided as a convenience
        because nobody wants to remember the above numbers. Negative numbers
        are considered 'prime children', which is described in the BIP32 spec
        as a leading 1 in a 32 bit unsigned int.
        This derivation is fully described at
        https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#child-key-derivation-functions  # nopep8
        """
        boundary = 0x80000000

        if abs(child_number) >= boundary:
            raise ValueError("Invalid child number %s" % child_number)

        # If is_prime isn't set, then we can infer it from the child_number
        if is_prime is None:
            # Prime children are either < 0 or > 0x80000000
            if child_number < 0:
                child_number = abs(child_number)
                is_prime = True
            elif child_number >= boundary:
                child_number -= boundary
                is_prime = True
            else:
                is_prime = False
        else:
            # Otherwise is_prime is set so the child_number should be between
            # 0 and 0x80000000
            if child_number < 0 or child_number >= boundary:
                raise ValueError(
                    "Invalid child number. Must be between 0 and %s" %
                    boundary)

        if not self.private_key and is_prime:
            raise ValueError(
                "Cannot compute a prime child without a private key")

        if is_prime:
            # Even though we take child_number as an int < boundary, the
            # internal derivation needs it to be the larger number.
            child_number = child_number + boundary
        child_number_hex = long_to_hex(child_number, 8)

        if is_prime:
            # Let data = concat(0x00, self.key, child_number)
            data = b'00' + self.private_key.get_key()
        else:
            data = self.get_public_key_hex()
        data += child_number_hex

        # Compute a 64 Byte I that is the HMAC-SHA512, using self.chain_code
        # as the seed, and data as the message.
        I = hmac.new(
            unhexlify(self.chain_code),
            msg=unhexlify(data),
            digestmod=sha512).digest()
        # Split I into its 32 Byte components.
        I_L, I_R = I[:32], I[32:]

        if long_or_int(hexlify(I_L), 16) >= SECP256k1.order:
            raise InvalidPrivateKeyError("The derived key is too large.")

        c_i = hexlify(I_R)
        private_exponent = None
        public_pair = None
        if self.private_key:
            # Use private information for derivation
            # I_L is added to the current key's secret exponent (mod n), where
            # n is the order of the ECDSA curve in use.
            private_exponent = (
                (long_or_int(hexlify(I_L), 16) +
                 long_or_int(self.private_key.get_key(), 16))
                % SECP256k1.order)
            # I_R is the child's chain code
        else:
            # Only use public information for this derivation
            g = SECP256k1.generator
            I_L_long = long_or_int(hexlify(I_L), 16)
            point = (_ECDSA_Public_key(g, g * I_L_long).point +
                     self.public_key.to_point())
            # I_R is the child's chain code
            public_pair = PublicPair(point.x(), point.y())

        child = self.__class__(
            chain_code=c_i,
            depth=self.depth + 1,  # we have to go deeper...
            parent_fingerprint=self.fingerprint,
            child_number=child_number_hex,
            private_exponent=private_exponent,
            public_pair=public_pair,
            network=self.network)
        if child.public_key.to_point() == INFINITY:
            raise InfinityPointException("The point at infinity is invalid.")
        if not as_private:
            return child.public_copy()
        return child

    def public_copy(self):
        """Clone this wallet and strip it of its private information."""
        return self.__class__(
            chain_code=self.chain_code,
            depth=self.depth,
            parent_fingerprint=self.parent_fingerprint,
            child_number=self.child_number,
            public_pair=self.public_key.to_public_pair(),
            network=self.network)

    def crack_private_key(self, child_private_key):
        """Crack the parent private key given a child private key.
        BIP32 has a vulnerability/feature that allows you to recover the
        master private key if you're given a master public key and any of its
        publicly-derived child private keys. This is a pretty serious security
        vulnerability that looks as innocuous as this:
        >>> w = Wallet.new_random_wallet()
        >>> child = w.get_child(0, is_prime=False)
        >>> w_pub = w.public_copy()
        >>> assert w_pub.private_key is None
        >>> master_public_key = w_pub.serialize_b58(private=False)
        >>> # Now you put master_public_key on your website
        >>> # and give somebody a private key
        >>> public_master = Wallet.deserialize(master_public_key)
        >>> cracked_private_master = public_master.crack_private_key(child)
        >>> assert w == cracked_private_master  # :(
        Implementation details from http://bitcoinmagazine.com/8396/deterministic-wallets-advantages-flaw/  # nopep8
        """
        if self.private_key:
            raise AssertionError("You already know the private key")
        if child_private_key.parent_fingerprint != self.fingerprint:
            raise ValueError("This is not a valid child")
        if child_private_key.child_number >= 0x80000000:
            raise ValueError(
                "Cannot crack private keys from private derivation")

        # Duplicate the public child derivation
        child_number_hex = long_to_hex(child_private_key.child_number, 8)
        data = self.get_public_key_hex() + child_number_hex
        I = hmac.new(
            unhexlify(self.chain_code),
            msg=unhexlify(data),
            digestmod=sha512).digest()
        I_L, I_R = I[:32], I[32:]
        # Public derivation is the same as private derivation plus some offset
        # knowing the child's private key allows us to find this offset just
        # by subtracting the child's private key from the parent I_L data
        privkey = PrivateKey(long_or_int(hexlify(I_L), 16),
                             network=self.network)
        parent_private_key = child_private_key.private_key - privkey
        return self.__class__(
            chain_code=self.chain_code,
            depth=self.depth,
            parent_fingerprint=self.parent_fingerprint,
            child_number=self.child_number,
            private_key=parent_private_key,
            network=self.network)

    def export_to_wif(self):
        """Export a key to WIF.
        See https://en.bitcoin.it/wiki/Wallet_import_format for a full
        description.
        """
        # Add the network byte, creating the "extended key"
        extended_key_hex = self.private_key.get_extended_key()
        # BIP32 wallets have a trailing \01 byte
        extended_key_bytes = unhexlify(extended_key_hex) + b'\01'
        # And return the base58-encoded result with a checksum
        return base58.b58encode_check(extended_key_bytes)

    def serialize(self, private=True):
        """Serialize this key.
        :param private: Whether or not the serialized key should contain
            private information. Set to False for a public-only representation
            that cannot spend funds but can create children. You want
            private=False if you are, for example, running an e-commerce
            website and want to accept bitcoin payments. See the README
            for more information.
        :type private: bool, defaults to True
        See the spec in `deserialize` for more details.
        """
        if private and not self.private_key:
            raise ValueError("Cannot serialize a public key as private")

        if private:
            network_version = long_to_hex(
                self.network.EXT_SECRET_KEY, 8)
        else:
            network_version = long_to_hex(
                self.network.EXT_PUBLIC_KEY, 8)
        depth = long_to_hex(self.depth, 2)
        parent_fingerprint = self.parent_fingerprint[2:]  # strip leading 0x
        child_number = long_to_hex(self.child_number, 8)
        chain_code = self.chain_code
        ret = (network_version + depth + parent_fingerprint + child_number +
               chain_code)
        # Private and public serializations are slightly different
        if private:
            ret += b'00' + self.private_key.get_key()
        else:
            ret += self.get_public_key_hex(compressed=True)
        return ensure_bytes(ret.lower())

    def serialize_b58(self, private=True):
        """Encode the serialized node in base58."""
        return ensure_str(
            base58.b58encode_check(unhexlify(self.serialize(private))))

    def to_address(self):
        """Create a public address from this Wallet.
        Public addresses can accept payments.
        https://en.bitcoin.it/wiki/Technical_background_of_Bitcoin_addresses
        """
        key = unhexlify(self.get_public_key_hex())
        # First get the hash160 of the key
        hash160_bytes = hash160(key)
        # Prepend the network address byte
        network_hash160_bytes = \
            chr_py2(self.network.PUBKEY_ADDRESS) + hash160_bytes
        # Return a base58 encoded address with a checksum
        return ensure_str(base58.b58encode_check(network_hash160_bytes))

    @classmethod #@memoize
    def deserialize(cls, key, network="bitcoin_testnet"):
        """Load the ExtendedBip32Key from a hex key.
        The key consists of
            * 4 byte version bytes (network key)
            * 1 byte depth:
                - 0x00 for master nodes,
                - 0x01 for level-1 descendants, ....
            * 4 byte fingerprint of the parent's key (0x00000000 if master key)
            * 4 byte child number. This is the number i in x_i = x_{par}/i,
              with x_i the key being serialized. This is encoded in MSB order.
              (0x00000000 if master key)
            * 32 bytes: the chain code
            * 33 bytes: the public key or private key data
              (0x02 + X or 0x03 + X for public keys, 0x00 + k for private keys)
              (Note that this also supports 0x04 + X + Y uncompressed points,
              but this is totally non-standard and this library won't even
              generate such data.)
        """
        network = Wallet.get_network(network)

        if len(key) in [78, (78 + 32)]:
            # we have a byte array, so pass
            pass
        else:
            key = ensure_bytes(key)
            if len(key) in [78 * 2, (78 + 32) * 2]:
                # we have a hexlified non-base58 key, continue!
                key = unhexlify(key)
            elif len(key) == 111:
                # We have a base58 encoded string
                key = base58.b58decode_check(key)
        # Now that we double checkd the values, convert back to bytes because
        # they're easier to slice
        version, depth, parent_fingerprint, child, chain_code, key_data = (
            key[:4], key[4], key[5:9], key[9:13], key[13:45], key[45:])

        version_long = long_or_int(hexlify(version), 16)
        exponent = None
        pubkey = None
        point_type = key_data[0]
        if not isinstance(point_type, six.integer_types):
            point_type = ord(point_type)
        if point_type == 0:
            # Private key
            if version_long != network.EXT_SECRET_KEY:
                raise incompatible_network_exception_factory(
                    network.NAME, network.EXT_SECRET_KEY,
                    version)
            exponent = key_data[1:]
        elif point_type in [2, 3, 4]:
            # Compressed public coordinates
            if version_long != network.EXT_PUBLIC_KEY:
                raise incompatible_network_exception_factory(
                    network.NAME, network.EXT_PUBLIC_KEY,
                    version)
            pubkey = PublicKey.from_hex_key(key_data, network=network)
            # Even though this was generated from a compressed pubkey, we
            # want to store it as an uncompressed pubkey
            pubkey.compressed = False
        else:
            raise ValueError("Invalid key_data prefix, got %s" % point_type)

        def l(byte_seq):
            if byte_seq is None:
                return byte_seq
            elif isinstance(byte_seq, six.integer_types):
                return byte_seq
            return long_or_int(hexlify(byte_seq), 16)

        return cls(depth=l(depth),
                   parent_fingerprint=l(parent_fingerprint),
                   child_number=l(child),
                   chain_code=l(chain_code),
                   private_exponent=l(exponent),
                   public_key=pubkey,
                   network=network)

    @classmethod
    def from_master_secret(cls, seed, network='monero'):
        """Generate a new PrivateKey from a secret key.
        :param seed: The key to use to generate this wallet. It may be a long
            string. Do not use a phrase from a book or song, as that will
            be guessed and is not secure. My advice is to not supply this
            argument and let me generate a new random key for you.
        :param network:
        See https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki#Serialization_format  # nopep8
        """
        network = Wallet.get_network(network)
        seed = ensure_bytes(seed)
        # Given a seed S of at least 128 bits, but 256 is advised
        # Calculate I = HMAC-SHA512(key="Bitcoin seed", msg=S)
        I = hmac.new(b"Bitcoin seed", msg=seed, digestmod=sha512).digest()
        # Split I into two 32-byte sequences, IL and IR.
        I_L, I_R = I[:32], I[32:]
        # Use IL as master secret key, and IR as master chain code.
        return cls(private_exponent=long_or_int(hexlify(I_L), 16),
                   chain_code=long_or_int(hexlify(I_R), 16),
                   network=network,
                   seed_secret=seed)

    @classmethod
    def from_master_secret_slow(cls, password, network=BitcoinMainNet):
        """
        Generate a new key from a password using 50,000 rounds of HMAC-SHA256.
        This should generate the same result as bip32.org.
        WARNING: The security of this method has not been evaluated.
        """
        # Make sure the password string is bytes
        key = ensure_bytes(password)
        data = unhexlify(b"0" * 64)  # 256-bit 0
        for i in range(50000):
            data = hmac.new(key, msg=data, digestmod=sha256).digest()
        return cls.from_master_secret(data, network)

    def __eq__(self, other):
        attrs = [
            'chain_code',
            'depth',
            'parent_fingerprint',
            'child_number',
            'private_key',
            'public_key',
            'network',
        ]
        return other and all(
            getattr(self, attr) == getattr(other, attr) for attr in attrs)

    def __ne__(self, other):
        return not self == other

    __hash__ = object.__hash__

    @classmethod
    def get_network(self, network):
        # returns a network class object

        response = None
        if network == "bitcoin_testnet" or network == "BTCTEST":
            response = BitcoinTestNet
        elif network == "bitcoin" or network == "BTC":
            response = BitcoinMainNet
        elif network == "dogecoin" or network == "DOGE":
            response = DogecoinMainNet
        elif network == "dogecoin_testnet" or network == "DOGETEST":
            response = DogecoinTestNet
        elif network == "litecoin" or network == "LTC":
            response = LitecoinMainNet
        elif network == "litecoin_testnet" or network == "LTCTEST":
            response = LitecoinTestNet
        elif network == "bitcoin_cash" or network == "BCH":
            response = BitcoinCashMainNet
        elif network == "bitcoin_gold" or network == "BTG":
            response = BitcoinGoldMainNet
        elif network == "dash" or network == "DASH":
            response = DashMainNet
        elif network == 'dash_testnet' or network == 'DASHTEST':
            response = DashTestNet
        elif network == 'monero':
            response = MoneroMainNet
        else:
            response = network
        return response

    @classmethod
    def new_random_wallet(cls, user_entropy=None, network=BitcoinMainNet):
        """
        Generate a new wallet using a randomly generated 512 bit seed.
        Args:
            user_entropy: Optional user-supplied entropy which is combined
                combined with the random seed, to help counteract compromised
                PRNGs.
        You are encouraged to add an optional `user_entropy` string to protect
        against a compromised CSPRNG. This will be combined with the output
        from the CSPRNG. Note that if you do supply this value it only adds
        additional entropy and will not be sufficient to recover the random
        wallet. If you're even saving `user_entropy` at all, you're doing it
        wrong.
        """
        seed = crypto.random_bytes(64)
        if user_entropy:
            user_entropy = str(user_entropy)  # allow for int/long
            seed += user_entropy
        return cls.from_master_secret(seed, network=network)


class InvalidPathError(Exception):
    pass


class InsufficientKeyDataError(ValueError):
    pass


class InvalidPrivateKeyError(ValueError):
    pass


class InvalidPublicKeyError(ValueError):
    pass


class KeyMismatchError(ValueError):
    pass


class InfinityPointException(Exception):
    pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import sys
import operator


# Useful for very coarse version differentiation.
PY3 = sys.version_info[0] == 3

if PY3:
    indexbytes = operator.getitem
    intlist2bytes = bytes
    int2byte = operator.methodcaller('to_bytes', 1, 'big')

else:
    int2byte = chr
    range = xrange

    def indexbytes(buf, i):
        return ord(buf[i])

    def intlist2bytes(l):
        return b"".join(chr(c) for c in l)





import ctypes as ct


FE = ct.c_uint32 * 10
MODM = ct.c_uint32 * 9
KEY_BUFF = ct.c_byte * 32


class Ge25519(ct.Structure):
    _fields_ = [
        ('x', FE),
        ('y', FE),
        ('z', FE),
        ('t', FE),
    ]


class Ge25519_p1p1(ct.Structure):
    _fields_ = [
        ('x', FE),
        ('y', FE),
        ('z', FE),
        ('t', FE),
    ]


class Ge25519_niels(ct.Structure):
    _fields_ = [
        ('ysubx', FE),
        ('xaddy', FE),
        ('t2d', FE),
    ]


class Ge25519_pniels(ct.Structure):  # projective Cached / Deuis
    _fields_ = [
        ('ysubx', FE),
        ('xaddy', FE),
        ('z', FE),
        ('t2d', FE),
    ]


XmrAmount = ct.c_uint64

# XmrKey is not a structure but a direct buffer so it can be used
# in the transaction in a straightforward way.
XmrKey = KEY_BUFF


class XmrCtKey(ct.Structure):
    _fields_ = [
        ('dest', XmrKey),
        ('mask', XmrKey),
    ]


XmrKey64 = XmrKey * 64


class XmrBoroSig(ct.Structure):
    _fields_ = [
        ('s0', XmrKey64),
        ('s1', XmrKey64),
        ('ee', XmrKey),
    ]


class XmrRangeSig(ct.Structure):
    _fields_ = [
        ('asig', XmrBoroSig),
        ('Ci', XmrKey64),
    ]


# ./sha2.h: 56
class struct__SHA256_CTX(ct.Structure):
    __slots__ = ['state', 'bitcount', 'buffer']
    _fields_ = [
        ('state', ct.c_uint32 * 8),
        ('bitcount', ct.c_uint64),
        ('buffer', ct.c_uint32 * (64 // ct.sizeof(ct.c_uint32))),
    ]


SHA256_CTX = struct__SHA256_CTX  # ./sha2.h: 56


# ./sha3.h: 60
class struct_SHA3_CTX(ct.Structure):
    __slots__ = ['hash', 'message', 'rest', 'block_size']

    _fields_ = [
        ('hash', ct.c_uint64 * 25),
        ('message', ct.c_uint64 * 24),
        ('rest', ct.c_uint),
        ('block_size', ct.c_uint),
    ]


SHA3_CTX = struct_SHA3_CTX


# ./blake256.h: 45
class struct_anon_2(ct.Structure):
    __slots__ = ['h', 's', 't', 'buflen', 'nullt', 'buf']
    _fields_ = [
        ('h', ct.c_uint32 * 8),
        ('s', ct.c_uint32 * 4),
        ('t', ct.c_uint32 * 2),
        ('buflen', ct.c_size_t),
        ('nullt', ct.c_uint8),
        ('buf', ct.c_ubyte * 64),
    ]


BLAKE256_CTX = struct_anon_2  # ./blake256.h: 45


# ./groestl.h: 54
class union_anon_3(ct.Union):
    __slots__ = ['wide', 'narrow']
    _fields_ = [
        ('wide', ct.c_uint64 * 16),
        ('narrow', ct.c_uint32 * 32),
    ]


# ./groestl.h: 59
class struct_anon_4(ct.Structure):
    __slots__ = ['buf', 'ptr', 'state', 'count']
    _fields_ = [
        ('buf', ct.c_ubyte * 128),
        ('ptr', ct.c_size_t),
        ('state', union_anon_3),
        ('count', ct.c_uint64),
    ]


sph_groestl_big_context = struct_anon_4  # ./groestl.h: 59
GROESTL512_CTX = sph_groestl_big_context # ./groestl.h: 61


class UnionHasher(ct.Union):
    __slots__ = ['sha2', 'sha3', 'blake', 'groestl']
    _fields_ = [
        ('sha2', SHA256_CTX),
        ('sha3', SHA3_CTX),
        ('blake', BLAKE256_CTX),
        ('groestl', GROESTL512_CTX),
    ]


# ./hasher.h: 61
class struct_anon_7(ct.Structure):
    __slots__ = ['type', 'ctx']
    _fields_ = [
        ('type', ct.c_int),
        ('ctx', UnionHasher),
    ]


Hasher = struct_anon_7





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import unittest
import binascii

import aiounittest

from monero_glue.xmr import monero, ring_ct, crypto


class RingCtTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(RingCtTest, self).__init__(*args, **kwargs)

    def test_range_proof(self):
        proof = ring_ct.prove_range(0)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)
        self.assertTrue(crypto.point_eq(proof[0], crypto.point_add(
            crypto.scalarmult_base(proof[1]),
            crypto.scalarmult_h(0))))

        proof = ring_ct.prove_range(0, mem_opt=False)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)
        self.assertTrue(crypto.point_eq(proof[0], crypto.point_add(
            crypto.scalarmult_base(proof[1]),
            crypto.scalarmult_h(0))))

    def test_range_proof_back(self):
        proof = ring_ct.prove_range(0, backend_impl=True)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

    def test_range_proof2(self):
        amount = 17 + (1 << 60)
        proof = ring_ct.prove_range(amount)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)
        self.assertTrue(crypto.point_eq(proof[0], crypto.point_add(
            crypto.scalarmult_base(proof[1]),
            crypto.scalarmult_h(amount))))

        proof = ring_ct.prove_range(amount, mem_opt=False, decode=True)
        res = ring_ct.ver_range(proof[0], proof[2], decode=False)
        self.assertTrue(res)

        res = ring_ct.ver_range(crypto.point_add(proof[0], crypto.scalarmult_base(crypto.sc_init(4))), proof[2], decode=False)
        self.assertFalse(res)

    def test_range_proof2_back(self):
        proof = ring_ct.prove_range(123456789, backend_impl=True)
        res = ring_ct.ver_range(proof[0], proof[2])
        self.assertTrue(res)

        res = ring_ct.ver_range(crypto.point_add(proof[0], crypto.scalarmult_base(crypto.sc_init(4))), proof[2])
        self.assertFalse(res)

    def test_range_proof3(self):
        proof = ring_ct.prove_range(123456789)
        rsig = proof[2]

        monero.recode_rangesig(rsig, encode=False)
        monero.recode_rangesig(rsig, encode=True)
        res = ring_ct.ver_range(proof[0], rsig)
        self.assertTrue(res)

    def test_range_proof_old(self):
        proof = ring_ct.prove_range(0, use_asnl=True, mem_opt=False, decode=True)
        res = ring_ct.ver_range(proof[0], proof[2], use_asnl=True, decode=False)
        self.assertTrue(res)

    def test_range_proof2_old(self):
        proof = ring_ct.prove_range(123456789, use_asnl=True, mem_opt=False, decode=True)
        res = ring_ct.ver_range(proof[0], proof[2], use_asnl=True, decode=False)
        self.assertTrue(res)
        res = ring_ct.ver_range(crypto.point_add(proof[0], crypto.scalarmult_base(crypto.sc_init(4))), proof[2], use_asnl=True, decode=False)
        self.assertFalse(res)

    def test_key_image_signature(self):
        ki = binascii.unhexlify(b'a248206cea806a7d60ea936cdc35efdf44a189b1026c4e658f42216aec155383')
        c0 = binascii.unhexlify(b'032725822d2c0f37bb67f29e116dc8c64ec02c4e8c69b147f596e7dbbc899409')
        r0 = binascii.unhexlify(b'2e4839b81c74f5b17d842d5f15723813b5287cfbc44539c9154d9222b65d2b07')
        pub = binascii.unhexlify(b'346354ca120bf3210976b1f5a9cee897628f3745fb80d5525c22f8cffd78a5c7')

        self.assertEqual(1, ring_ct.check_ring_singature(ki, crypto.decodepoint(ki),
                                                         [crypto.decodepoint(pub)],
                                                         [[crypto.decodeint(c0), crypto.decodeint(r0)]]))


if __name__ == "__main__":
    unittest.main()  # pragma: no cover





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Dusan Klinec, ph4r05, 2018

import logging
from . import ec_picker


logger = logging.getLogger(__name__)
backend = ec_picker.get_ec_backend()


if backend == ec_picker.EC_BACKEND_PY:
    from monero_glue.xmr.core.ec_py import *

elif backend == ec_picker.EC_BACKEND_TREZOR:
    try:
        from monero_glue.xmr.core.ec_trezor import *
        
    except Exception as e:
        logger.warning('Trezor-crypto backend not usable: %s' % e)
        if ec_picker.get_ec_backend_force():
            raise

        from monero_glue.xmr.core.ec_py import *

else:
    raise ValueError('Unknown EC backend: %s' % backend)




import ctypes as ct
import os
from . import trezor_types as tt
from .trezor_types import *


# Loaded library instance
CLIB = None


def open_lib(lib_path=None, try_env=True):
    """
    Opens the library
    :param lib_path:
    :param try_env:
    :return:
    """
    global CLIB

    # Open Trezor-crypto
    if lib_path is None and try_env:
        lib_path = os.getenv('LIBTREZOR_CRYPTO_PATH', None)
    if lib_path is None:
        lib_path = os.path.join(os.path.dirname(__file__), './libtrezor-crypto.so')
    if lib_path is None or not os.path.exists(lib_path):
        raise FileNotFoundError('Trezor-Crypto lib not found')

    CLIB = ct.cdll.LoadLibrary(lib_path)

    # Functions
    CLIB.random_init.restype = ct.c_int
    CLIB.random_buffer.argtypes = [ct.c_void_p, ct.c_size_t]
    CLIB.random_uniform.argtypes = [ct.c_uint32]
    CLIB.random_uniform.restype = ct.c_uint32
    CLIB.random32.restype = ct.c_uint32

    #
    # MODM / scalar values
    #

    CLIB.contract256_modm.argtypes = [tt.KEY_BUFF, tt.MODM]
    CLIB.expand256_modm.argtypes = [tt.MODM, ct.c_void_p, ct.c_size_t]
    CLIB.add256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM]
    CLIB.neg256_modm.argtypes = [tt.MODM, tt.MODM]
    CLIB.sub256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM]
    CLIB.mul256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM]
    CLIB.reduce256_modm.argtypes = [tt.MODM]
    CLIB.barrett_reduce256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM]
    CLIB.set256_modm.argtypes = [tt.MODM, ct.c_uint64]
    CLIB.get256_modm.argtypes = [ct.POINTER(ct.c_uint64), tt.MODM]
    CLIB.get256_modm.restype = ct.c_int

    CLIB.eq256_modm.argtypes = [tt.MODM, tt.MODM]
    CLIB.eq256_modm.restype = ct.c_int

    CLIB.cmp256_modm.argtypes = [tt.MODM, tt.MODM]
    CLIB.cmp256_modm.restype = ct.c_int

    CLIB.iszero256_modm.argtypes = [tt.MODM]
    CLIB.iszero256_modm.restype = ct.c_int

    CLIB.copy256_modm.argtypes = [tt.MODM, tt.MODM]

    CLIB.check256_modm.argtypes = [tt.MODM]
    CLIB.check256_modm.restype = ct.c_int

    CLIB.mulsub256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM, tt.MODM]
    CLIB.muladd256_modm.argtypes = [tt.MODM, tt.MODM, tt.MODM, tt.MODM]

    #
    # FE
    #

    CLIB.curve25519_copy.argtypes = [tt.FE, tt.FE]
    CLIB.curve25519_add.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_add_after_basic.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_add_reduce.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_sub.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_scalar_product.argtypes = [tt.FE, tt.FE, ct.c_uint32]
    CLIB.curve25519_sub_after_basic.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_sub_reduce.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_neg.argtypes = [tt.FE, tt.FE]
    CLIB.curve25519_mul.argtypes = [tt.FE, tt.FE, tt.FE]
    CLIB.curve25519_square.argtypes = [tt.FE, tt.FE]
    CLIB.curve25519_square_times.argtypes = [tt.FE, tt.FE, ct.c_int]
    CLIB.curve25519_expand.argtypes = [tt.FE, tt.KEY_BUFF]
    CLIB.curve25519_contract.argtypes = [tt.KEY_BUFF, tt.FE]
    CLIB.curve25519_recip.argtypes = [tt.FE, tt.FE]

    CLIB.curve25519_set.argtypes = [tt.FE, ct.c_uint32]
    CLIB.curve25519_isnegative.argtypes = [tt.FE]
    CLIB.curve25519_isnonzero.argtypes = [tt.FE]
    CLIB.curve25519_reduce.argtypes = [tt.FE, tt.FE]
    CLIB.curve25519_expand_reduce.argtypes = [tt.FE, tt.KEY_BUFF]

    CLIB.curve25519_isnegative.restype = ct.c_int
    CLIB.curve25519_isnonzero.restype = ct.c_int

    #
    # GE
    #

    CLIB.ed25519_verify.argtypes = [ct.c_char_p, ct.c_char_p, ct.c_size_t]
    CLIB.ge25519_p1p1_to_partial.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_p1p1)]
    CLIB.ge25519_p1p1_to_full.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_p1p1)]
    CLIB.ge25519_full_to_pniels.argtypes = [ct.POINTER(tt.Ge25519_pniels), ct.POINTER(tt.Ge25519)]

    CLIB.ge25519_double_p1p1.argtypes = [ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_nielsadd2_p1p1.argtypes = [ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_niels), ct.c_char]
    CLIB.ge25519_pnielsadd_p1p1.argtypes = [ct.POINTER(tt.Ge25519_p1p1), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_pniels), ct.c_char]
    CLIB.ge25519_double_partial.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_double.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_nielsadd2.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_niels)]
    CLIB.ge25519_pnielsadd.argtypes = [ct.POINTER(tt.Ge25519_pniels), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519_pniels)]
    CLIB.ge25519_pack.argtypes = [tt.KEY_BUFF, ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_unpack_negative_vartime.argtypes = [ct.POINTER(tt.Ge25519), tt.KEY_BUFF]
    CLIB.ge25519_set_neutral.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_double_scalarmult_vartime.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM, tt.MODM]
    CLIB.ge25519_double_scalarmult_vartime2.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519), tt.MODM]
    CLIB.ge25519_scalarmult.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM]
    CLIB.ge25519_check.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_fromfe_check.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_eq.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_copy.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_set_base.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_mul8.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_neg_partial.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_neg_full.argtypes = [ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_reduce.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_norm.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519)]
    CLIB.ge25519_add.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.c_char]
    CLIB.ge25519_fromfe_frombytes_vartime.argtypes = [ct.POINTER(tt.Ge25519), tt.KEY_BUFF]
    CLIB.ge25519_unpack_vartime.argtypes = [ct.POINTER(tt.Ge25519), tt.KEY_BUFF]
    CLIB.ge25519_scalarmult_base_wrapper.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM]
    CLIB.ge25519_scalarmult_wrapper.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM]
    CLIB.ge25519_set_xmr_h.argtypes = [ct.POINTER(tt.Ge25519)]

    CLIB.ed25519_verify.restype = ct.c_int
    CLIB.ge25519_unpack_negative_vartime.restype = ct.c_int
    CLIB.ge25519_check.restype = ct.c_int
    CLIB.ge25519_fromfe_check.restype = ct.c_int
    CLIB.ge25519_eq.restype = ct.c_int
    CLIB.ge25519_unpack_vartime.restype = ct.c_int

    #
    # XMR
    #

    CLIB.xmr_random_scalar.argtypes = [tt.MODM]
    CLIB.xmr_fast_hash.argtypes = [tt.KEY_BUFF, ct.c_void_p, ct.c_size_t]

    CLIB.xmr_hasher_init.argtypes = [ct.POINTER(tt.Hasher)]
    CLIB.xmr_hasher_update.argtypes = [ct.POINTER(tt.Hasher), ct.c_void_p, ct.c_size_t]
    CLIB.xmr_hasher_final.argtypes = [ct.POINTER(tt.Hasher), tt.KEY_BUFF]
    CLIB.xmr_hasher_copy.argtypes = [ct.POINTER(tt.Hasher), ct.POINTER(tt.Hasher)]

    CLIB.xmr_hash_to_scalar.argtypes = [tt.MODM, ct.c_void_p, ct.c_size_t]
    CLIB.xmr_hash_to_ec.argtypes = [ct.POINTER(tt.Ge25519), ct.c_void_p, ct.c_size_t]
    CLIB.xmr_derivation_to_scalar.argtypes = [tt.MODM, ct.POINTER(tt.Ge25519), ct.c_uint32]
    CLIB.xmr_generate_key_derivation.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), tt.MODM]
    CLIB.xmr_derive_private_key.argtypes = [tt.MODM, ct.POINTER(tt.Ge25519), ct.c_uint32, tt.MODM]
    CLIB.xmr_derive_public_key.argtypes = [ct.POINTER(tt.Ge25519), ct.POINTER(tt.Ge25519), ct.c_uint32, ct.POINTER(tt.Ge25519)]
    CLIB.xmr_gen_c.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM, ct.c_uint64]
    CLIB.xmr_add_keys2.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM, tt.MODM, ct.POINTER(tt.Ge25519)]
    CLIB.xmr_add_keys2_vartime.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM, tt.MODM, ct.POINTER(tt.Ge25519)]
    CLIB.xmr_add_keys3.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519)]
    CLIB.xmr_add_keys3_vartime.argtypes = [ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519), tt.MODM, ct.POINTER(tt.Ge25519)]
    CLIB.xmr_get_subaddress_secret_key.argtypes = [tt.MODM, ct.c_uint32, ct.c_uint32, tt.MODM]
    CLIB.xmr_gen_range_sig.argtypes = [ct.POINTER(tt.XmrRangeSig), ct.POINTER(tt.Ge25519), tt.MODM, tt.XmrAmount, ct.POINTER(tt.MODM)]

    init_lib()

#
# Wrappers
#

def init_lib():
    """
    Initializes Trezor crypto library
    :return:
    """
    res = CLIB.random_init()
    if res < 0:
        raise ValueError('Library initialization error: %s' % res)
    return res


def random_buffer(sz):
    buff = (ct.c_uint8 * sz)()
    CLIB.random_buffer(ct.byref(buff), sz)
    return bytes(buff)

random_buffer_r = random_buffer


#
# SC
#


def expand256_modm(r, buff):
    CLIB.expand256_modm(r, buff, len(buff))
    return r


def contract256_modm(buff, sc):
    CLIB.contract256_modm(buff, sc)
    return bytes(buff)


def add256_modm(r, a, b):
    CLIB.add256_modm(r, a, b)
    return r


def neg256_modm(r, a):
    CLIB.neg256_modm(r, a)
    return r


def sub256_modm(r, a, b):
    CLIB.sub256_modm(r, a, b)
    return r


def mul256_modm(r, a, b):
    CLIB.mul256_modm(r, a, b)
    return r


def reduce256_modm(a):
    CLIB.reduce256_modm(a)
    return a


def barrett_reduce256_modm(r, a, b):
    CLIB.barrett_reduce256_modm(r, a, b)
    return r


def set256_modm(a, b):
    CLIB.set256_modm(a, ct.c_uint64(b))


def get256_modm(a, b):
    return CLIB.get256_modm(a, b)


def get256_modm_r(a):
    r = ct.c_uint64()
    res = CLIB.get256_modm(ct.byref(r), a)
    if not res:
        raise ValueError('Get256_modm failed')
    return r.value


def init256_modm(r, a):
    CLIB.set256_modm(r, ct.c_uint64(a))
    return r


def eq256_modm(a, b):
    return CLIB.eq256_modm(a, b)


def cmp256_modm(a, b):
    return CLIB.cmp256_modm(a, b)


def iszero256_modm(a):
    return CLIB.iszero256_modm(a)


def copy256_modm(a, b):
    return CLIB.copy256_modm(a, b)


def clone256_modm(a):
    r = tt.MODM()
    CLIB.copy256_modm(r, a)
    return r


def check256_modm(a):
    return CLIB.check256_modm(a)


def mulsub256_modm(r, a, b, c):
    CLIB.mulsub256_modm(r, a, b, c)
    return r


def muladd256_modm(r, a, b, c):
    CLIB.muladd256_modm(r, a, b, c)
    return r


def expand256_modm_r(buff):
    m = tt.MODM()
    CLIB.expand256_modm(m, bytes(buff), len(buff))
    return m


def contract256_modm_r(sc):
    buff = tt.KEY_BUFF()
    CLIB.contract256_modm(buff, sc)
    return bytes(buff)


def add256_modm_r(a, b):
    r = tt.MODM()
    CLIB.add256_modm(r, a, b)
    return r


def neg256_modm_r(a):
    r = tt.MODM()
    CLIB.neg256_modm(r, a)
    return r


def sub256_modm_r(a, b):
    r = tt.MODM()
    CLIB.sub256_modm(r, a, b)
    return r


def mul256_modm_r(a, b):
    r = tt.MODM()
    CLIB.mul256_modm(r, a, b)
    return r


def barrett_reduce256_modm_r(a, b):
    r = tt.MODM()
    CLIB.barrett_reduce256_modm(r, a, b)
    return r


def init256_modm_r(a):
    r = tt.MODM()
    CLIB.set256_modm(r, ct.c_uint64(a))
    return r


def mulsub256_modm_r(a, b, c):
    r = tt.MODM()
    CLIB.mulsub256_modm(r, a, b, c)
    return r


def muladd256_modm_r(a, b, c):
    r = tt.MODM()
    CLIB.muladd256_modm(r, a, b, c)
    return r


#
# FE
#


def curve25519_clone(a):
    r = tt.Ge25519()
    CLIB.curve25519_copy(r, a)
    return r


def curve25519_copy(r, a):
    return CLIB.curve25519_copy(r, a)


def curve25519_copy_r(a):
    r = tt.FE()
    CLIB.curve25519_copy(r, a)
    return r


def curve25519_add(r, a, b):
    return CLIB.curve25519_add(r, a, b)


def curve25519_add_r(a, b):
    r = tt.FE()
    CLIB.curve25519_add(r, a, b)
    return r


def curve25519_add_after_basic(r, a, b):
    return CLIB.curve25519_add_after_basic(r, a, b)


def curve25519_add_after_basic_r(a, b):
    r = tt.FE()
    CLIB.curve25519_add_after_basic(r, a, b)
    return r


def curve25519_add_reduce(r, a, b):
    return CLIB.curve25519_add_reduce(r, a, b)


def curve25519_add_reduce_r(a, b):
    r = tt.FE()
    CLIB.curve25519_add_reduce(r, a, b)
    return r


def curve25519_sub(r, a, b):
    return CLIB.curve25519_sub(r, a, b)


def curve25519_sub_r(a, b):
    r = tt.FE()
    CLIB.curve25519_sub(r, a, b)
    return r


def curve25519_scalar_product(r, a, b):
    return CLIB.curve25519_scalar_product(r, a, b)


def curve25519_scalar_product_r(a, b):
    r = tt.FE()
    CLIB.curve25519_scalar_product(r, a, b)
    return r


def curve25519_sub_after_basic(r, a, b):
    return CLIB.curve25519_sub_after_basic(r, a, b)


def curve25519_sub_after_basic_r(a, b):
    r = tt.FE()
    CLIB.curve25519_sub_after_basic(r, a, b)
    return r


def curve25519_sub_reduce(r, a, b):
    return CLIB.curve25519_sub_reduce(r, a, b)


def curve25519_sub_reduce_r(a, b):
    r = tt.FE()
    CLIB.curve25519_sub_reduce(r, a, b)
    return r


def curve25519_neg(r, a):
    return CLIB.curve25519_neg(r, a)


def curve25519_neg_r(a):
    r = tt.FE()
    CLIB.curve25519_neg(r, a)
    return r


def curve25519_mul(r, a, b):
    return CLIB.curve25519_mul(r, a, b)


def curve25519_mul_r(a, b):
    r = tt.FE()
    CLIB.curve25519_mul(r, a, b)
    return r


def curve25519_square(r, a):
    return CLIB.curve25519_square(r, a)


def curve25519_square_r(a):
    r = tt.FE()
    CLIB.curve25519_square(r, a)
    return r


def curve25519_square_times(r, a, b):
    return CLIB.curve25519_square_times(r, a, b)


def curve25519_square_times_r(a, b):
    r = tt.FE()
    CLIB.curve25519_square_times(r, a, b)
    return r


def curve25519_expand(r, a):
    a = tt.KEY_BUFF(*a)
    return CLIB.curve25519_expand(r, a)


def curve25519_expand_r(a):
    r = tt.FE()
    a = tt.KEY_BUFF(*a)
    CLIB.curve25519_expand(r, a)
    return r


def curve25519_contract(r, a):
    return CLIB.curve25519_contract(r, a)


def curve25519_contract_r(a):
    r = tt.KEY_BUFF()
    CLIB.curve25519_contract(r, a)
    return bytes(r)


def curve25519_recip(r, a):
    return CLIB.curve25519_recip(r, a)


def curve25519_recip_r(a):
    r = tt.FE()
    CLIB.curve25519_recip(r, a)
    return r


def curve25519_set(r, a):
    return CLIB.curve25519_set(r, a)


def curve25519_set_r(a):
    r = tt.FE()
    CLIB.curve25519_set(r, a)
    return r


def curve25519_set_d(r):
    return CLIB.curve25519_set_d(r)


def curve25519_set_d_r():
    r = tt.FE()
    CLIB.curve25519_set_d(r)
    return r


def curve25519_isnegative(a):
    return CLIB.curve25519_isnegative(a)


def curve25519_isnonzero(a):
    return CLIB.curve25519_isnonzero(a)


def curve25519_reduce(r, a):
    return CLIB.curve25519_reduce(r, a)


def curve25519_reduce_r(a):
    r = tt.FE()
    CLIB.curve25519_reduce(r, a)
    return r


def curve25519_expand_reduce(r, a):
    return CLIB.curve25519_expand_reduce(r, a)


def curve25519_expand_reduce_r(a):
    r = tt.FE()
    CLIB.curve25519_expand_reduce(r, a)
    return r


#
# GE
#


def ed25519_verify(a, b, c):
    return CLIB.ed25519_verify(a, b, c)


def ge25519_p1p1_to_partial(r, a):
    return CLIB.ge25519_p1p1_to_partial(ct.byref(r), ct.byref(a))


def ge25519_p1p1_to_partial_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_p1p1_to_partial(ct.byref(r), ct.byref(a))
    return r


def ge25519_p1p1_to_full(r, a):
    return CLIB.ge25519_p1p1_to_full(ct.byref(r), ct.byref(a))


def ge25519_p1p1_to_full_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_p1p1_to_full(ct.byref(r), ct.byref(a))
    return r


def ge25519_full_to_pniels(r, a):
    return CLIB.ge25519_full_to_pniels(ct.byref(r), ct.byref(a))


def ge25519_full_to_pniels_r(a):
    r = tt.Ge25519_pniels()
    CLIB.ge25519_full_to_pniels(ct.byref(r), ct.byref(a))
    return r


def ge25519_double_p1p1(r, a):
    return CLIB.ge25519_double_p1p1(ct.byref(r), ct.byref(a))


def ge25519_double_p1p1_r(a):
    r = tt.Ge25519_p1p1()
    CLIB.ge25519_double_p1p1(ct.byref(r), ct.byref(a))
    return r


def ge25519_nielsadd2_p1p1(r, a, b, c):
    return CLIB.ge25519_nielsadd2_p1p1(ct.byref(r), ct.byref(a), ct.byref(b), c)


def ge25519_nielsadd2_p1p1_r(a, b, c):
    r = tt.Ge25519_p1p1()
    CLIB.ge25519_nielsadd2_p1p1(ct.byref(r), ct.byref(a), ct.byref(b), c)
    return r


def ge25519_pnielsadd_p1p1(r, a, b, c):
    return CLIB.ge25519_pnielsadd_p1p1(ct.byref(r), ct.byref(a), ct.byref(b), c)


def ge25519_pnielsadd_p1p1_r(a, b, c):
    r = tt.Ge25519_p1p1()
    CLIB.ge25519_pnielsadd_p1p1(ct.byref(r), ct.byref(a), ct.byref(b), c)
    return r


def ge25519_double_partial(r, a):
    return CLIB.ge25519_double_partial(ct.byref(r), ct.byref(a))


def ge25519_double_partial_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_double_partial(ct.byref(r), ct.byref(a))
    return r


def ge25519_double(r, a):
    return CLIB.ge25519_double(ct.byref(r), ct.byref(a))


def ge25519_double_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_double(ct.byref(r), ct.byref(a))
    return r


def ge25519_nielsadd2(r, a):
    return CLIB.ge25519_nielsadd2(ct.byref(r), ct.byref(a))


def ge25519_nielsadd2_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_nielsadd2(ct.byref(r), ct.byref(a))
    return r


def ge25519_pnielsadd(r, a, b):
    return CLIB.ge25519_pnielsadd(ct.byref(r), ct.byref(a), ct.byref(b))


def ge25519_pnielsadd_r(a, b):
    r = tt.Ge25519_pniels()
    CLIB.ge25519_pnielsadd(ct.byref(r), ct.byref(a), ct.byref(b))
    return r


def ge25519_pack(r, a):
    return CLIB.ge25519_pack(r, ct.byref(a))


def ge25519_pack_r(a):
    r = tt.KEY_BUFF()
    CLIB.ge25519_pack(r, ct.byref(a))
    return bytes(r)


def ge25519_unpack_negative_vartime(a, b):
    b = tt.KEY_BUFF(*b)
    return CLIB.ge25519_unpack_negative_vartime(ct.byref(a), b)


def ge25519_set_neutral(r):
    return CLIB.ge25519_set_neutral(ct.byref(r))


def ge25519_set_neutral_r():
    r = tt.Ge25519()
    CLIB.ge25519_set_neutral(ct.byref(r))
    return r


def ge25519_double_scalarmult_vartime(r, a, b, c):
    return CLIB.ge25519_double_scalarmult_vartime(ct.byref(r), ct.byref(a), b, c)


def ge25519_double_scalarmult_vartime_r(a, b, c):
    r = tt.Ge25519()
    CLIB.ge25519_double_scalarmult_vartime(ct.byref(r), ct.byref(a), b, c)
    return r


def ge25519_double_scalarmult_vartime2(r, a, b, c, d):
    return CLIB.ge25519_double_scalarmult_vartime2(ct.byref(r), ct.byref(a), b, ct.byref(c), d)


def ge25519_double_scalarmult_vartime2_r(a, b, c, d):
    r = tt.Ge25519()
    CLIB.ge25519_double_scalarmult_vartime2(ct.byref(r), ct.byref(a), b, ct.byref(c), d)
    return r


def ge25519_scalarmult(r, a, b):
    return CLIB.ge25519_scalarmult(ct.byref(r), ct.byref(a), b)


def ge25519_scalarmult_r(a, b):
    r = tt.Ge25519()
    CLIB.ge25519_scalarmult(ct.byref(r), ct.byref(a), b)
    return r


def ge25519_check(a):
    return CLIB.ge25519_check(ct.byref(a))


def ge25519_fromfe_check(a):
    return CLIB.ge25519_fromfe_check(ct.byref(a))


def ge25519_eq(a, b):
    return CLIB.ge25519_eq(ct.byref(a), ct.byref(b))


def ge25519_copy(r, a):
    return CLIB.ge25519_copy(ct.byref(r), ct.byref(a))


def ge25519_copy_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_copy(ct.byref(r), ct.byref(a))
    return r


def ge25519_set_base(r):
    return CLIB.ge25519_set_base(ct.byref(r))


def ge25519_set_base_r():
    r = tt.Ge25519()
    CLIB.ge25519_set_base(ct.byref(r))
    return r


def ge25519_mul8(r, a):
    return CLIB.ge25519_mul8(ct.byref(r), ct.byref(a))


def ge25519_mul8_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_mul8(ct.byref(r), ct.byref(a))
    return r


def ge25519_neg_partial(r):
    return CLIB.ge25519_neg_partial(ct.byref(r))


def ge25519_neg_partial_r():
    r = tt.Ge25519()
    CLIB.ge25519_neg_partial(ct.byref(r))
    return r


def ge25519_neg_full(r):
    return CLIB.ge25519_neg_full(ct.byref(r))


def ge25519_neg_full_r():
    r = tt.Ge25519()
    CLIB.ge25519_neg_full(ct.byref(r))
    return r


def ge25519_reduce(r, a):
    return CLIB.ge25519_reduce(ct.byref(r), ct.byref(a))


def ge25519_reduce_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_reduce(ct.byref(r), ct.byref(a))
    return r


def ge25519_norm(r, a):
    return CLIB.ge25519_norm(ct.byref(r), ct.byref(a))


def ge25519_norm_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_norm(ct.byref(r), ct.byref(a))
    return r


def ge25519_add(r, a, b, c):
    return CLIB.ge25519_add(ct.byref(r), ct.byref(a), ct.byref(b), c)


def ge25519_add_r(a, b, c):
    r = tt.Ge25519()
    CLIB.ge25519_add(ct.byref(r), ct.byref(a), ct.byref(b), c)
    return r


def ge25519_fromfe_frombytes_vartime(r, a):
    a = tt.KEY_BUFF(*a)
    return CLIB.ge25519_fromfe_frombytes_vartime(ct.byref(r), a)


def ge25519_fromfe_frombytes_vartime_r(a):
    a = tt.KEY_BUFF(*a)
    r = tt.Ge25519()
    CLIB.ge25519_fromfe_frombytes_vartime(ct.byref(r), a)
    return r


def ge25519_unpack_vartime(a, b):
    b = tt.KEY_BUFF(*b)
    return CLIB.ge25519_unpack_vartime(ct.byref(a), b)


def ge25519_unpack_vartime_r(buff):
    pt = tt.Ge25519()
    #buff = tt.KEY_BUFF(*buff)
    r = CLIB.ge25519_unpack_vartime(ct.byref(pt), buff)
    if r != 1:
        raise ValueError('Point decoding error')
    return pt


def ge25519_scalarmult_base_wrapper(r, a):
    return CLIB.ge25519_scalarmult_base_wrapper(ct.byref(r), a)


def ge25519_scalarmult_base_wrapper_r(a):
    r = tt.Ge25519()
    CLIB.ge25519_scalarmult_base_wrapper(ct.byref(r), a)
    return r


def ge25519_scalarmult_wrapper(r, a, b):
    return CLIB.ge25519_scalarmult_wrapper(ct.byref(r), ct.byref(a), b)


def ge25519_scalarmult_wrapper_r(a, b):
    r = tt.Ge25519()
    CLIB.ge25519_scalarmult_wrapper(ct.byref(r), ct.byref(a), b)
    return r


def ge25519_set_xmr_h(r):
    return CLIB.ge25519_set_xmr_h(ct.byref(r))


def ge25519_set_xmr_h_r():
    r = tt.Ge25519()
    CLIB.ge25519_set_xmr_h(ct.byref(r))
    return r


def xmr_random_scalar(r):
    return CLIB.xmr_random_scalar(r)


def xmr_random_scalar_r():
    r = tt.MODM()
    CLIB.xmr_random_scalar(r)
    return r


def xmr_fast_hash(r, a, b):
    return CLIB.xmr_fast_hash(r, bytes(a), b)


def xmr_fast_hash_r(a):
    r = tt.KEY_BUFF()
    CLIB.xmr_fast_hash(r, bytes(a), len(a))
    return bytes(r)


def xmr_hasher_init_r():
    h = tt.Hasher()
    CLIB.xmr_hasher_init(ct.byref(h))
    return h


def xmr_hasher_update(h, buff):
    CLIB.xmr_hasher_update(ct.byref(h), bytes(buff), len(buff))


def xmr_hasher_final_r(h):
    r = tt.KEY_BUFF()
    CLIB.xmr_hasher_final(ct.byref(h), r)
    return bytes(r)


def xmr_hasher_copy_r(h):
    hd = tt.Hasher()
    CLIB.xmr_hasher_copy(ct.byref(hd), ct.byref(h))
    return hd


def xmr_hash_to_scalar(r, a):
    return CLIB.xmr_hash_to_scalar(r, bytes(a), len(a))


def xmr_hash_to_scalar_r(a):
    r = tt.MODM()
    CLIB.xmr_hash_to_scalar(r, bytes(a), len(a))
    return r


def xmr_hash_to_ec(r, a):
    return CLIB.xmr_hash_to_ec(ct.byref(r), bytes(a), len(a))


def xmr_hash_to_ec_r(a):
    r = tt.Ge25519()
    CLIB.xmr_hash_to_ec(ct.byref(r), bytes(a), len(a))
    return r


def xmr_derivation_to_scalar(r, a, b):
    return CLIB.xmr_derivation_to_scalar(r, ct.byref(a), b)


def xmr_derivation_to_scalar_r(a, b):
    r = tt.MODM()
    CLIB.xmr_derivation_to_scalar(r, ct.byref(a), b)
    return r


def xmr_generate_key_derivation(r, a, b):
    return CLIB.xmr_generate_key_derivation(ct.byref(r), ct.byref(a), b)


def xmr_generate_key_derivation_r(a, b):
    r = tt.Ge25519()
    CLIB.xmr_generate_key_derivation(ct.byref(r), ct.byref(a), b)
    return r


def xmr_derive_private_key(r, a, b, c):
    return CLIB.xmr_derive_private_key(r, ct.byref(a), b, c)


def xmr_derive_private_key_r(a, b, c):
    r = tt.MODM()
    CLIB.xmr_derive_private_key(r, ct.byref(a), b, c)
    return r


def xmr_derive_public_key(r, a, b, c):
    return CLIB.xmr_derive_public_key(ct.byref(r), ct.byref(a), b, ct.byref(c))


def xmr_derive_public_key_r(a, b, c):
    r = tt.Ge25519()
    CLIB.xmr_derive_public_key(ct.byref(r), ct.byref(a), b, ct.byref(c))
    return r


def xmr_gen_c(r, a, b):
    return CLIB.xmr_gen_c(ct.byref(r), a, b)


def xmr_gen_c_r(a, b):
    r = tt.Ge25519()
    CLIB.xmr_gen_c(ct.byref(r), a, b)
    return r


def xmr_add_keys2(r, a, b, c):
    return CLIB.xmr_add_keys2(ct.byref(r), a, b, ct.byref(c))


def xmr_add_keys2_r(a, b, c):
    r = tt.Ge25519()
    CLIB.xmr_add_keys2(ct.byref(r), a, b, ct.byref(c))
    return r


def xmr_add_keys2_vartime(r, a, b, c):
    return CLIB.xmr_add_keys2_vartime(ct.byref(r), a, b, ct.byref(c))


def xmr_add_keys2_vartime_r(a, b, c):
    r = tt.Ge25519()
    CLIB.xmr_add_keys2_vartime(ct.byref(r), a, b, ct.byref(c))
    return r


def xmr_add_keys3(r, a, b, c, d):
    return CLIB.xmr_add_keys3(ct.byref(r), a, ct.byref(b), c, ct.byref(d))


def xmr_add_keys3_r(a, b, c, d):
    r = tt.Ge25519()
    CLIB.xmr_add_keys3(ct.byref(r), a, ct.byref(b), c, ct.byref(d))
    return r


def xmr_add_keys3_vartime(r, a, b, c, d):
    return CLIB.xmr_add_keys3_vartime(ct.byref(r), a, ct.byref(b), c, ct.byref(d))


def xmr_add_keys3_vartime_r(a, b, c, d):
    r = tt.Ge25519()
    CLIB.xmr_add_keys3_vartime(ct.byref(r), a, ct.byref(b), c, ct.byref(d))
    return r


def xmr_get_subaddress_secret_key(r, a, b, c):
    return CLIB.xmr_get_subaddress_secret_key(r, a, b, c)


def xmr_get_subaddress_secret_key_r(a, b, c):
    r = tt.MODM()
    CLIB.xmr_get_subaddress_secret_key(r, a, b, c)
    return r


#
# XMR
#


def gen_range_proof(amount, last_mask):
    """
    Trezor crypto range proof
    :param amount:
    :param last_mask:
    :return:
    """
    rsig = tt.XmrRangeSig()
    C = tt.Ge25519()
    mask = tt.MODM()
    last_mask_ptr = ct.byref(last_mask) if last_mask else None

    CLIB.xmr_gen_range_sig(ct.byref(rsig), ct.byref(C), mask, amount, last_mask_ptr)

    return C, mask, rsig

